"""Copy records from the first KGTK file to the output file,
expanding one column (usualy node2) into seperate columns for each subfield.

TODO: Need KgtkWriterOptions
"""

from argparse import _MutuallyExclusiveGroup, Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'aliases': ['denormalize_node2'],
        'help': 'Copy a KGTK file, exploding one column (usualy node2) into seperate columns for each subfield.',
        'description': 'Copy a KGTK file, exploding one column (usually node2) into seperate columns for each subfield. ' +
        'If a cell in the column being exploded contains a list, that record is optionally expanded into multiple records ' +
        'before explosion, with all other columns copied-as is.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert explode --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalue import KgtkValueFields
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument(      "--column", dest="column_name", help="The name of the column to explode. (default=%(default)s).", default=KgtkFormat.NODE2)

    fgroup: _MutuallyExclusiveGroup = parser.add_mutually_exclusive_group()
    fgroup.add_argument(      "--types", dest="type_names", nargs='*',
                              help="The KGTK data types for which fields should be exploded. (default=%(default)s).",
                              choices=KgtkFormat.DataType.choices(),
                              default=KgtkFormat.DataType.choices())

    fgroup.add_argument(      "--fields", dest="field_names", nargs='*', 
                              help=h("The names of the fields to extract (overrides --types). (default=%(default)s)."),
                              choices=KgtkValueFields.FIELD_NAMES)

    parser.add_argument(      "--prefix", dest="prefix", help="The prefix for exploded column names. (default=%(default)s).",
                              default=KgtkFormat.NODE2 + ";" + KgtkFormat.KGTK_NAMESPACE)

    parser.add_argument(      "--overwrite", dest="overwrite_columns", metavar="True|False",
                              help="Indicate that it is OK to overwrite existing columns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--expand", dest="expand_list", metavar="True|False",
                              help="When True, expand source cells that contain a lists, else fail if a source cell contains a list. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--show-data-types", dest="show_data_types", metavar="True|False",
                              help="Print the list of data types and exit. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--show-field-names", dest="show_field_names", metavar="True|False",
                              help="Print the list of field names and exit. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--show-field-formats", dest="show_field_formats", metavar="True|False",
                              help="Print the list of field names and formats, then exit. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file:KGTKFiles,
        column_name: str,
        type_names: typing.List[str],
        field_names: typing.List[str],
        prefix: str,
        overwrite_columns: bool,
        expand_list: bool,
        show_data_types: bool,
        show_field_names: bool,
        show_field_formats: bool,
        output_format: typing.Optional[str],
        
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkexplode import KgtkExplode
    from kgtk.value.kgtkvalue import KgtkValueFields
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        print("--column %s" % column_name, file=error_file, flush=True)
        print("--prefix %s" % prefix, file=error_file, flush=True)
        print("--overwrite %s" % str(overwrite_columns), file=error_file, flush=True)
        print("--expand %s" % str(expand_list), file=error_file, flush=True)
        if type_names is not None:
            print("--types %s" % " ".join(type_names), file=error_file, flush=True)
        if field_names is not None:
            print("--fields %s" % " ".join(field_names), file=error_file, flush=True)
        print("--show-data-types %s" % str(show_data_types), file=error_file, flush=True)
        print("--show-field-names %s" % str(show_field_names), file=error_file, flush=True)
        print("--show-field-formats %s" % str(show_field_formats), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format=%s" % output_format, file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if show_data_types:
        data_type: str
        for data_type in KgtkFormat.DataType.choices():
            print("%s" % data_type, file=error_file, flush=True)
        return 0

    if show_field_names:
        field_name: str
        for field_name in sorted(KgtkValueFields.FIELD_NAMES):
            print("%s" % field_name, file=error_file, flush=True)
        return 0

    if show_field_formats:
        field_name: str
        for field_name in sorted(KgtkValueFields.FIELD_NAME_FORMATS.keys()):
            field_format: str = KgtkValueFields.FIELD_NAME_FORMATS[field_name]
            print("| %20s | %-5s |" % (field_name, field_format), file=error_file, flush=True)
        return 0

    try:
        ex: KgtkExplode = KgtkExplode(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            output_format=output_format,
            column_name=column_name,
            type_names=type_names,
            field_names=field_names,
            prefix=prefix,
            overwrite_columns=overwrite_columns,
            expand_list=expand_list,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""Copy records from the first KGTK file to the output file,
compacting lists.

The list compacting algorithm requires that input records with the same keyset
be grouped (not necessarily sorted).  kgtkcompact will sort its input
in-memory by default, but that can be disabled for large input files using an
external presorter.

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=False)
class KgtkCompact(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    output_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    key_column_names: typing.List[str] = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                         iterable_validator=attr.validators.instance_of(list)))
    compact_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    deduplicate: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    # The field separator used in multifield joins.  The KGTK list character should be safe.
    field_separator: str = attr.ib(validator=attr.validators.instance_of(str), default=KgtkFormat.KEY_FIELD_SEPARATOR)

    sorted_input: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    verify_sort: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    lists_in_input: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    build_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    idbuilder_options: typing.Optional[KgtkIdBuilderOptions] = attr.ib(default=None)

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # We unfroze this object to keep these rwitable fields around across
    # invocations of process_row.
    #
    # TODO: Introduce a row processing object?
    output_line_count: int = 0
    current_key: typing.Optional[str] = None
    current_row: typing.Optional[typing.List[str]] = None
    current_row_lists: typing.Optional[typing.List[typing.Optional[typing.List[str]]]] = None

    FIELD_SEPARATOR_DEFAULT: str = KgtkFormat.KEY_FIELD_SEPARATOR

    def build_key(self, row: typing.List[str], key_columns: typing.List[int])->str:
        key: str = ""
        idx: int
        first: bool = True
        for idx in key_columns:
            if first:
                first = False
            else:
                key += self.field_separator
            key += row[idx]
        return key

    def compact_row(self):
        if self.current_row_lists is None:
            return

        # Preallocate the list, this might be more efficient than appending to it..
        self.current_row = [None] * len(self.current_row_lists)
        idx: int
        item_list: typing.Optional[typing.List[str]]
        for idx, item_list in enumerate(self.current_row_lists):
            if item_list is None:
                self.current_row[idx] = ""
            else:
                # We don't need to use KgtkValue.join_unique_list(item_list)
                # because self.merge_row(...) and self.expand_row(...) ensure that
                # there are no duplicates.
                #
                # TODO: run timing studies to determine which approach is more efficient.
                self.current_row[idx] = KgtkValue.join_sorted_list(item_list)
        self.current_row_lists = None

    def expand_row(self, row: typing.List[str], force: bool = False):
        if not self.lists_in_input and not force:
            self.current_row = row # Optimization: leave the row alone if possible.
            return
        
        # Preallocate the list, this might be more efficient than appending to it..
        self.current_row_lists = [None] * len(row)
        idx: int
        item: str
        for idx, item in enumerate(row):
            if len(item) == 0:
                continue # Ignore empty items.

            # Start the new current item list:
            current_item_list: typing.Optional[typing.List[str]] = None

            # The row item might itself be a list.
            item2: str
            for item2 in KgtkValue.split_list(item):
                if len(item2) == 0:
                    continue # Ignore empty items
                
                if current_item_list is None:
                    # This is the first item.
                    current_item_list = [ item2 ]
                    continue
                
                # There might be duplicate items in the row item's list.
                if item2 not in current_item_list:
                    current_item_list.append(item2) # Add unique items.
                    
            self.current_row_lists[idx] = current_item_list

    def merge_row(self,  row: typing.List[str]):
        if self.current_row_lists is None:
            if self.current_row is None:
                # TODO: raise a better error
                raise ValueError("Inconsistent state 31 in merge_row.")
            else:
                # We deferred expanding the previous row, but we must do so
                # now:
                self.expand_row(self.current_row, force=True)
                if self.current_row_lists is None:
                    # Keep mypy happy by ensuring that self.current_row_lists is not None.
                    #
                    # TODO: raise a better error.
                    raise ValueError("Inconsistent state #2 in merge_row.")

        idx: int
        item: str
        for idx, item in enumerate(row):
            if len(item) == 0:
                continue # Ignore empty items

            # We will modify the current item list in place!
            current_item_list: typing.Optional[typing.List[str]] = self.current_row_lists[idx]

            # The row item might itself be a list.
            item2: str
            for item2 in KgtkValue.split_list(item):
                if len(item2) == 0:
                    continue # Ignore empty items.

                if current_item_list is None:
                    # This is the first item.
                    current_item_list = [ item2 ]
                    self.current_row_lists[idx] = current_item_list
                    continue

                # There might be duplicate items in the row item's list.
                if item2 not in current_item_list:
                    current_item_list.append(item2) # Add unique items.

    # TODO: Create an optimized version of this routine without the very verbose debugginng messages.
    def process_row(self,
                    input_key: str,
                    row: typing.List[str],
                    line_number: int,
                    idb: typing.Optional[KgtkIdBuilder],
                    ew: KgtkWriter,
                    flush: bool = False):
        if self.very_verbose:
            print("Input key %s" % repr(input_key), file=self.error_file, flush=True)
        # Note:  This code makes the assumption that row lengths do not vary!
        if self.current_key is not None:
            if self.very_verbose:
                print("No current key", file=self.error_file, flush=True)
            # We have a record being built.  Write it?
            if flush or self.current_key != input_key:
                if self.very_verbose:
                    if flush:
                        print("flush", file=self.error_file, flush=True)
                    else:
                        print("current_key %s != input_key %s" % (repr(self.current_key), repr(input_key)), file=self.error_file, flush=True)
                # self.current_key != input_key means that the key is changing.
                self.compact_row()
                if self.current_row is not None:
                    if self.very_verbose:
                        print("writing %s" % repr(self.field_separator.join(self.current_row)), file=self.error_file, flush=True)
                    if idb is None:
                        ew.write(self.current_row)
                    else:
                        ew.write(idb.build(self.current_row, line_number))
                    self.output_line_count += 1
                self.current_key = None
                self.current_row = None

        if flush:
            # This was a flush request.  We're done.
            return

        # Are we starting a new key?
        if self.current_key is None:
            # Save the new row.
            if self.very_verbose:
                print("New current_key %s" % repr(self.current_key), file=self.error_file, flush=True)
            self.current_key = input_key
            if self.very_verbose:
                print("Expand row %s" % self.field_separator.join(row), file=self.error_file, flush=True)
            self.expand_row(row)
        else:
            # Merge into an existing row.
            if self.very_verbose:
                print("Merge row", file=self.error_file, flush=True)
            self.merge_row(row)

    def process(self):

        # Open the input file.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        # If requested, creat the ID column builder.
        # Assemble the list of output column names.
        output_column_names: typing.List[str]
        idb: typing.Optional[KgtkIdBuilder] = None
        if self.build_id:
            if self.idbuilder_options is None:
                raise ValueError("ID build requested but ID builder options are missing")
            idb = KgtkIdBuilder.new(kr, self.idbuilder_options)
            output_column_names = idb.column_names
        else:
            output_column_names = kr.column_names

        # Build the list of key column edges:
        key_idx_list: typing.List[int] = [ ]

        if self.deduplicate:
            # Use the columns in the order the appear in the input file.  If you want to
            # use --presorted, then the input needs to be sorted the same way.
            key_idx_list = range(kr.column_count)
            
        elif len(self.key_column_names) == 0:
            if kr.is_edge_file:
                # Add the KGTK edge file required columns.
                key_idx_list.append(kr.node1_column_idx)
                key_idx_list.append(kr.label_column_idx)
                key_idx_list.append(kr.node2_column_idx)
                if not self.compact_id and kr.id_column_idx >= 0:
                    key_idx_list.append(kr.id_column_idx)

            elif kr.is_node_file:
                # Add the KGTK node file required column:
                key_idx_list.append(kr.id_column_idx)

            else:
                raise ValueError("The input file is neither an edge nor a node file.  Key columns must be supplied.")

        else:
            # Append additional columns to the list of key column indices,
            # silently removing duplicates, but complaining about unknown names.
            #
            # TODO: warn about duplicates?
            column_name: str
            for column_name in self.key_column_names:
                if column_name not in kr.column_name_map:
                    raise ValueError("Column %s is not in the input file" % (column_name))
                key_idx: int = kr.column_name_map[column_name]
                if key_idx not in key_idx_list:
                    key_idx_list.append(key_idx)

        if self.verbose:
            key_idx_list_str: typing.List[str] = [ ]
            for key_idx in key_idx_list:
                key_idx_list_str.append(str(key_idx))
            print("key indexes: %s" % " ".join(key_idx_list_str), file=self.error_file, flush=True)
            
        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         use_mgzip=self.reader_options.use_mgzip, # Hack!
                                         mgzip_threads=self.reader_options.mgzip_threads, # Hack!
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)        
        input_line_count: int = 0
        row: typing.List[str] = [ ]
        input_key: str
        prev_input_key: typing.Optional[str] = None
        going_up: typing.Optional[bool] = None
        if self.sorted_input:
            if self.verbose:
                print("Reading the input data from %s" % self.input_file_path, file=self.error_file, flush=True)
            for row in kr:
                input_line_count += 1
                input_key = self.build_key(row, key_idx_list)
                if self.verify_sort:
                    if prev_input_key is None:
                        prev_input_key = input_key
                    else:
                        if going_up is None:
                            if prev_input_key < input_key:
                                going_up = True
                                prev_input_key = input_key
                            elif prev_input_key > input_key:
                                going_up = False
                                prev_input_key = input_key
                            else:
                                pass # No change in input key
                        elif going_up:
                            if prev_input_key < input_key:
                                prev_input_key = input_key
                            elif prev_input_key > input_key:
                                raise ValueError("Line %d sort violation going up: prev='%s' curr='%s'" % (input_line_count,
                                                                                                           prev_input_key.replace(self.field_separator, KgtkFormat.LIST_SEPARATOR),
                                                                                                           input_key.replace(self.field_separator, KgtkFormat.LIST_SEPARATOR)))
                            else:
                                pass # No change in input_key
                        else:
                            if prev_input_key > input_key:
                                prev_input_key = input_key
                            elif prev_input_key < input_key:
                                raise ValueError("Line %d sort violation going down: prev='%s' curr='%s'" % (input_line_count,
                                                                                                             prev_input_key.replace(self.field_separator, KgtkFormat.LIST_SEPARATOR),
                                                                                                             input_key.replace(self.field_separator, KgtkFormat.LIST_SEPARATOR)))
                            else:
                                pass # No change in input_key
                            
                self.process_row(input_key, row, input_line_count, idb, ew)
            
        else:
            if self.verbose:
                print("Sorting the input data from %s" % self.input_file_path, file=self.error_file, flush=True)
            # Map key values to lists of input and output data.
            input_map: typing.MutableMapping[str, typing.List[typing.List[str]]] = { }

            for row in kr:
                input_line_count += 1
                input_key = self.build_key(row, key_idx_list)
                if input_key in input_map:
                    # Append the row to an existing list for that key.
                    input_map[input_key].append(row)
                else:
                    # Create a new list of rows for this key.
                    input_map[input_key] = [ row ]

            if self.verbose:
                print("Processing the sorted input data", file=self.error_file, flush=True)
            
            for input_key in sorted(input_map.keys()):
                for row in input_map[input_key]:
                    self.process_row(input_key, row, input_line_count, idb, ew)

        # Flush the final row, if any.  We pass the last row read for
        # feedback, such as an ID uniqueness violation.
        self.process_row("", row, input_line_count, idb, ew, flush=True)
        
        if self.verbose:
            print("Read %d records, wrote %d records." % (input_line_count, self.output_line_count), file=self.error_file, flush=True)
        
        ew.close()

def main():
    """
    Test the KGTK compact processor.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data (default=%(default)s)", type=Path, nargs="?", default="-")

    parser.add_argument(      "--columns", dest="key_column_names",
                              help="The key columns to identify records for compaction. " +
                              "(default=id for node files, (node1, label, node2, id) for edge files).", nargs='+', default=[ ])

    parser.add_argument(      "--compact-id", dest="compact_id",
                              help="Indicate that the ID column in KGTK edge files should be compacted. " +
                              "Normally, if the ID column exists, it is not compacted, " +
                              "as there are use cases that need to maintain distinct lists of secondary edges for each ID value. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--deduplicate", dest="deduplicate",
                              help="Treat all columns as key columns, overriding --columns and --compact-id. " +
                              "This will remove completely duplicate records without compacting any new lists. " +
                              "(default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--presorted", dest="sorted_input",
                              help="Indicate that the input has been presorted (or at least pregrouped). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--verify-sort", dest="verify_sort",
                              help="If the input has been presorted, verify its consistency (disable if only pregrouped). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--lists-in-input", dest="lists_in_input",
                              help="Assume that the input file may contain lists (disable when certain it does not). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--build-id", dest="build_id",
                              help="Build id values in an id column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkIdBuilderOptions.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_args(args)    
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("input: %s" % str(args.input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        print("--columns %s" % " ".join(args.key_column_names), file=error_file, flush=True)
        print("--compact-id=%s" % str(args.compact_id), file=error_file, flush=True)
        print("--deduplicate=%s" % str(args.deduplicate), file=error_file, flush=True)
        print("--presorted=%s" % str(args.sorted_input), file=error_file, flush=True)
        print("--verify-sort=%s" % str(args.verify_sort), file=error_file, flush=True)
        print("--lists-in-input=%s" % str(args.lists_in_input), file=error_file, flush=True)
        print("--build-id=%s" % str(args.build_id), file=error_file, flush=True)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kc: KgtkCompact = KgtkCompact(
        input_file_path=args.input_file_path,
        key_column_names=args.key_column_names,
        compact_id=args.compact_id,
        deduplicate=args.deduplicate,
        sorted_input=args.sorted_input,
        verify_sort=args.verify_sort,
        lists_in_input=args.lists_in_input,
        output_file_path=args.output_file_path,
        build_id=args.build_id,
        idbuilder_options=idbuilder_options,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    kc.process()

if __name__ == "__main__":
    main()



"""
Remove columns from a KGTK file.
"""
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Remove columns from a KGTK file.',
        'description': 'Remove specific columns from a KGTK file.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument('-c', "--columns", action="store", type=str, dest="columns", nargs='+', required=True,
                        help="Columns to remove as a comma- or space-separated strings, e.g., id,docid or id docid")

    parser.add_argument(      "--split-on-commas", dest="split_on_commas", help="When True, parse the list of columns, splitting on commas. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--split-on-spaces", dest="split_on_spaces", help="When True, parse the list of columns, splitting on spaces. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--strip-spaces", dest="strip_spaces", help="When True, parse the list of columns, stripping whitespace. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--all-except", dest="all_except", help="When True, remove all columns except the listed ones. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--ignore-missing-columns", dest="ignore_missing_columns", help="When True, ignore missing columns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, default_mode=KgtkReaderMode.NONE, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        columns: typing.Optional[typing.List[str]],

        split_on_commas: bool,
        split_on_spaces: bool,
        strip_spaces: bool,

        all_except: bool,
        ignore_missing_columns: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        if columns is not None:
            print("--columns=%s" % " ".join(columns), file=error_file)
        print("--split-on-commas=%s" % str(split_on_commas), file=error_file)
        print("--split-on-spaces=%s" % str(split_on_spaces), file=error_file)
        print("--strip-spaces=%s" % str(strip_spaces), file=error_file)
        print("--all-except=%s" % str(all_except), file=error_file)
        print("--ignore-missing-columns=%s" % str(ignore_missing_columns), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:

        if columns is None:
            columns = [ ] # This simplifies matters.

        if split_on_spaces:
            # We will be very lenient, and allow space-seperated arguments
            # *inside* shell quoting, e.g.
            #
            # kgtk remove_columns -c 'name name2 name3'
            #
            # Do not enable this option if spaces are legal inside your
            # column names.
            columns = " ".join(columns).split()
        remove_columns: typing.List[str] = [ ]
        arg: str
        column_name: str
        for arg in columns:
            if split_on_commas:
                for column_name in arg.split(","):
                    if strip_spaces:
                        column_name = column_name.strip()
                    if len(column_name) > 0:
                        remove_columns.append(column_name)
            else:
                if strip_spaces:
                    arg = arg.strip()
                if len(arg) > 0:
                    remove_columns.append(arg)
        if verbose:
            if all_except:
                print("Removing all columns except %d columns: %s" % (len(remove_columns), " ".join(remove_columns)), file=error_file, flush=True)
            else:
                print("Removing %d columns: %s" % (len(remove_columns), " ".join(remove_columns)), file=error_file, flush=True)
        if len(remove_columns) == 0:
            raise KGTKException("No columns to remove")

        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        output_column_names: typing.List[str]

        trouble_column_names: typing.List[str] = [ ]
        if all_except:
            if not ignore_missing_columns:
                for column_name in remove_columns:
                    if column_name not in kr.column_names:
                        print("Error: cannot retain unknown column '%s'." % column_name, file=error_file, flush=True)
                        trouble_column_names.append(column_name)

            output_column_names = [ ]
            for column_name in kr.column_names:
                if column_name in remove_columns:
                    output_column_names.append(column_name)
                
        else:
            output_column_names = kr.column_names.copy()
            for column_name in remove_columns:
                if column_name in output_column_names:
                    output_column_names.remove(column_name)

                elif not ignore_missing_columns:
                    print("Error: cannot remove unknown column '%s'." % column_name, file=error_file, flush=True)
                    trouble_column_names.append(column_name)

        if len(trouble_column_names) > 0:
            raise KGTKException("Unknown columns %s" % " ".join(trouble_column_names))

        if verbose:
            print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        shuffle_list: typing.List[int] = kw.build_shuffle_list(kr.column_names)

        input_line_count: int = 0
        row: typing.List[str]
        for row in kr:
            input_line_count += 1
            kw.write(row, shuffle_list=shuffle_list)

        if verbose:
            print("Processed %d rows." % (input_line_count), file=error_file, flush=True)

        kw.close()

        return 0

    except Exception as e:
        kgtk_exception_auto_handler(e)
        return 1



"""
Unreify RDF values in KGTK files
"""
from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.unreify.kgtksortbuffer import KgtkSortBuffer
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=False)
class KgtkUnreifyValues(KgtkFormat):
    # Note: If you change any of these default values, be sure to change the
    # explanation in the corresponding parser.add_argument(...) call.
    DEFAULT_TRIGGER_LABEL_VALUE: typing.Optional[str] = None
    DEFAULT_TRIGGER_NODE2_VALUE: typing.Optional[str] = None
    DEFAULT_VALUE_LABEL_VALUE: typing.Optional[str] = None
    DEFAULT_OLD_LABEL_VALUE: typing.Optional[str] = None
    DEFAULT_NEW_LABEL_VALUE: typing.Optional[str] = None
    DEFAULT_ALLOW_MULTIPLE_VALUES: bool = False
    DEFAULT_ALLOW_EXTRA_COLUMNS: bool = False

    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reified_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    unreified_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    uninvolved_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    trigger_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_TRIGGER_LABEL_VALUE)
    trigger_node2_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_TRIGGER_NODE2_VALUE)
    value_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_VALUE_LABEL_VALUE)
    old_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_OLD_LABEL_VALUE)
    new_label_value: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)),
                                                    default=DEFAULT_NEW_LABEL_VALUE)

    allow_multiple_values: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_MULTIPLE_VALUES)
    allow_extra_columns: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_EXTRA_COLUMNS)

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    output_format: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None) # TODO: use an enum

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # Working variables:
    output_line_count: int = attr.ib(default=0)

    def make_keygen(self, old_label_value: str)->KgtkSortBuffer.KEYGEN_TYPE:
        # Create a key generator function passing old_label_value as a
        # closure value.
        def keygen(buf: 'KgtkSortBuffer', row: typing.List[str])->str:
            label_column_idx: int = buf.label_column_idx
            if label_column_idx < 0:
                raise ValueError("Unknown label column.")
            
            if row[label_column_idx] == old_label_value:
                node2_column_idx: int = buf.node2_column_idx
                if node2_column_idx < 0:
                    raise ValueError("Unknown node2 column.")
                return row[node2_column_idx]

            else:
                node1_column_idx: int = buf.node1_column_idx
                if node1_column_idx < 0:
                    raise ValueError("Unknown node1 column.")
                return row[node1_column_idx]
        return keygen

    def process(self):
        # Open the input file.
        if self.verbose:
            print("Opening the input file: %s" % str(self.input_file_path), file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          mode=KgtkReaderMode.EDGE, # Must be an edge file.
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        output_column_names: typing.List[str] = kr.column_names.copy()

        node1_column_idx: int = kr.node1_column_idx
        node1_column_name: str = output_column_names[node1_column_idx]

        label_column_idx: int = kr.label_column_idx
        label_column_name: str = output_column_names[label_column_idx]

        node2_column_idx: int = kr.node2_column_idx
        node2_column_name: str = output_column_names[node2_column_idx]

        # Adding an ID column?
        new_id_column: bool = False
        id_column_idx: int = kr.id_column_idx
        if id_column_idx < 0:
            new_id_column = True
            id_column_idx = len(output_column_names)
            output_column_names.append(KgtkFormat.ID)
        id_column_name: str = output_column_names[id_column_idx]

        # There should be exactly 4 output columns, iincluding an ID column.
        # If there are additional columns, some content may be lost when
        # unreifying records.
        num_columns: int = len(output_column_names)
        if num_columns < 4 or (num_columns > 4 and not self.allow_extra_columns):
            raise ValueError("Expecting 4 output columns, found %d." % num_columns)

        if self.verbose:
            print("Opening the output file: %s" % str(self.output_file_path), file=self.error_file, flush=True)
        # Open the output file.
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         output_format=self.output_format,
                                         require_all_columns=not self.allow_extra_columns,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=self.allow_extra_columns,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        reifiedw: typing.Optional[KgtkWriter] = None
        if self.reified_file_path is not None:
            if self.verbose:
                print("Opening the reified value output file: %s" % str(self.reified_file_path), file=self.error_file, flush=True)
            reifiedw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                                   self.reified_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=not self.allow_extra_columns,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=self.allow_extra_columns,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        unreifiedw: typing.Optional[KgtkWriter] = None
        if self.unreified_file_path is not None:
            if self.verbose:
                print("Opening the unreified value output file: %s" % str(self.unreified_file_path), file=self.error_file, flush=True)
            unreifiedw: KgtkWriter = KgtkWriter.open(output_column_names,
                                                   self.unreified_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=True,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=False,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        uninvolvedw: typing.Optional[KgtkWriter] = None
        if self.uninvolved_file_path is not None:
            if self.verbose:
                print("Opening the uninvolved records output file: %s" % str(self.uninvolved_file_path), file=self.error_file, flush=True)
            uninvolvedw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                                   self.uninvolved_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=True,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=False,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        if self.verbose:
            print("Reading and grouping the input records.", file=self.error_file, flush=True)
        ksb: KgtkSortBuffer = KgtkSortBuffer.readall(kr, grouped=True, keygen=self.make_keygen(self.old_label_value))

        input_group_count: int = 0
        input_line_count: int = 0
        self.output_line_count = 0
        unreification_count: int = 0

        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        node1_group: typing.List[typing.List[str]]
        for node1_group in ksb.groupiterate():
            input_group_count += 1

            saw_error: bool = False
            saw_trigger: bool = False
            node1_value: typing.Optional[str] = None
            node2_values: typing.Set[str] = set()

            old_label_node2_value: typing.Optional[str] = None
            trigger_node1_value: typing.Optional[str] = None
           
            potential_edge_attributes: typing.List[typing.List[str]] = [ ]

            row: typing.List[str]
            for row in node1_group:
                input_line_count += 1
                node1: str = row[node1_column_idx]
                label: str = row[label_column_idx]
                node2: str = row[node2_column_idx]

                if label == self.trigger_label_value and node2 == self.trigger_node2_value:
                    if saw_trigger:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Duplicate trigger in input group %d (%s)" % (input_group_count, node1_value), file=self.error_file, flush=True)
                    saw_trigger = True
                    trigger_node1_value = node1

                elif label == self.value_label_value:
                    if len(node2_values) > 0 and node2 not in node2_values and not self.allow_multiple_values:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Multiple values in input group %d" % (input_group_count), file=self.error_file, flush=True)
                        saw_error = True
                    node2_values.add(node2)
                    
                elif label == self.old_label_value:
                    node1_value = node1
                    old_label_node2_value = node2

                else:
                    potential_edge_attributes.append(row)
                    
            if saw_trigger and \
               node1_value is not None and \
               len(node2_values) > 0 and \
               old_label_node2_value == trigger_node1_value and \
               not saw_error:
                # Unreification was triggered.
                unreification_count += 1

                node2_value: str = KgtkFormat.LIST_SEPARATOR.join(list(node2_values))

                if reifiedw is not None:
                    for row in node1_group:
                        reifiedw.write(row)

                self.write_new_edge(kw,
                                    reifiedw,
                                    potential_edge_attributes,
                                    node1_value,
                                    node2_value,
                                    trigger_node1_value,
                                    label_column_idx,
                                    node2_column_idx,
                                    node1_column_name,
                                    label_column_name,
                                    node2_column_name,
                                    id_column_name,
                )
                
            else:
                # Unreification was not triggered.  Pass this group of rows
                # through unchanged, except for possibly appending an ID
                # column.
                self.pass_group_through(kw, uninvolvedw, node1_group, new_id_column)

        if self.verbose:
            print("Processed %d records in %d groups." % (input_line_count, input_group_count), file=self.error_file, flush=True)
            print("Unreified %d groups." % unreification_count, file=self.error_file, flush=True)
            print("Wrote %d output records" % self.output_line_count, file=self.error_file, flush=True)

        
        kw.close()
        if reifiedw is not None:
            reifiedw.close()
        if unreifiedw is not None:
            unreifiedw.close()
        if uninvolvedw is not None:
            uninvolvedw.close()


    def make_new_id(self, edge_id: str, count: int, width: int)->str:
        # Generate a new ID that will sort after the new edge.
        # What if the existing ID is not a symbol or a string?
        #
        # TODO: Handle cases where the existing ID is not a symbol or string.
        new_id: str
        if edge_id.startswith(KgtkFormat.STRING_SIGIL) and edge_id.endswith(KgtkFormat.STRING_SIGIL):
            new_id = edge_id[:-1] + "-" + str(count).zfill(width) + KgtkFormat.STRING_SIGIL
        else:
            new_id = edge_id + "-" + str(count).zfill(width)
        return new_id

    def get_width(self, max_count: int)->int:
        return len(str(max_count).strip())
        
    def write_new_edge(self,
                       kw: KgtkWriter,
                       unreifiedw: typing.Optional[KgtkWriter],
                       potential_edge_attributes: typing.List[typing.List[str]],
                       node1_value: str,
                       node2_value: str,
                       edge_id: str,
                       label_column_idx: int,
                       node2_column_idx: int,
                       node1_column_name: str,
                       label_column_name: str,
                       node2_column_name: str,
                       id_column_name: str,
    ):
        new_label_value: str = self.new_label_value if self.new_label_value is not None else self.value_label_value
        kw.writemap({
            node1_column_name: node1_value,
            label_column_name: new_label_value,
            node2_column_name: node2_value,
            id_column_name: edge_id,
        })
        self.output_line_count += 1
        
        if unreifiedw is not None:
            unreifiedw.writemap({
                node1_column_name: node1_value,
                label_column_name: new_label_value,
                node2_column_name: node2_value,
                id_column_name: edge_id,
            })

        self.write_edge_attributes(kw,
                                   unreifiedw,
                                   potential_edge_attributes,
                                   edge_id,
                                   label_column_idx,
                                   node2_column_idx,
                                   node1_column_name,
                                   label_column_name,
                                   node2_column_name,
                                   id_column_name,
        )

    def write_edge_attributes(self,
                              kw: KgtkWriter,
                              unreifiedw: typing.Optional[KgtkWriter],
                              potential_edge_attributes: typing.List[typing.List[str]],
                              edge_id: str,
                              label_column_idx: int,
                              node2_column_idx: int,
                              node1_column_name: str,
                              label_column_name: str,
                              node2_column_name: str,
                              id_column_name: str,
    ):
        width: int = self.get_width(len(potential_edge_attributes))
        attribute_number: int = 0
        edge_row: typing.List[str]
        for edge_row in potential_edge_attributes:
            attribute_number += 1
            attr_edge_id: str = self.make_new_id(edge_id, attribute_number, width)

            kw.writemap({
                node1_column_name: edge_id,
                label_column_name: edge_row[label_column_idx],
                node2_column_name: edge_row[node2_column_idx],
                id_column_name: attr_edge_id
            })
            self.output_line_count += 1
                
            if unreifiedw is not None:
                unreifiedw.writemap({
                    node1_column_name: edge_id,
                    label_column_name: edge_row[label_column_idx],
                    node2_column_name: edge_row[node2_column_idx],
                    id_column_name: attr_edge_id
                })
                

    def pass_group_through(self,
                           kw: KgtkWriter,
                           uninvolvedw: typing.Optional[KgtkWriter],
                           node1_group: typing.List[typing.List[str]],
                           new_id_column: bool):
        # Unreification was not triggered.  Pass this group of rows
        # through unchanged, except for possibly appending an ID
        # column.
        #
        # TODO: Perhaps we'd like to build an ID value at the same time?
        row: typing.List[str]
        for row in node1_group:
            if uninvolvedw is not None:
                uninvolvedw.write(row)
                
            if new_id_column:
                row = row.copy()
                row.append("")

            kw.write(row)
            self.output_line_count += 1
        

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument(      "--trigger-label", dest="trigger_label_value", required=True,
                                  help="A value in the label (or its alias) column that identifies the trigger record. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_TRIGGER_LABEL_VALUE)

        parser.add_argument(      "--trigger-node2", dest="trigger_node2_value", required=True,
                                  help="A value in the node2 (or its alias) column that identifies the trigger record. " +
                                  "This is a required parameter for which there is no default value. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_TRIGGER_NODE2_VALUE)
    
        parser.add_argument(      "--value-label", dest="value_label_value", required=True,
                                  help="A value in the label (or its alias) column that identifies the record with the node2 value for the new, unreified edge. " +
                                  "This is a required parameter for which there is no default value. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_VALUE_LABEL_VALUE)
    
        parser.add_argument(      "--old-label", dest="old_label_value", required=True,
                                  help="A value in the label (or its alias) column  that identifies the edge with the node1 value being unreified. " +
                                  "The value in the node1 (or its alias) column of this record will be used in the node1 (or its alias) column for the " +
                                  "new, unreified edge. " +
                                  "This is a required parameter for which there is no default value. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_OLD_LABEL_VALUE)
    
        parser.add_argument(      "--new-label", dest="new_label_value",
                                  help="The value to be entered in the label (or its alias) column of the new, unreified edge. " +
                                  "If not specified (None), the value from --value-label is used. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_NEW_LABEL_VALUE)

        parser.add_argument(      "--allow-multiple-values", dest="allow_multiple_values",
                                  help="When true, allow multiple values (a '|' list) in the node2 (or its alias) column of the new, unreified edge. " +
                                  "(default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_MULTIPLE_VALUES)

        parser.add_argument(      "--allow-extra-columns", dest="allow_extra_columns",
                                  help="When true, allow extra columns (beyond node1, label, node2, and id, or their aliases. " +
                                  "Warning: the contents of these columns may be lost silently in unreified statements. " +
                                  "(default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_MULTIPLE_VALUES)

def main():
    """
    Test the KGTK copy template.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument("-i", "--input-file", dest="input_file_path",
                        help="The KGTK input file. (default=%(default)s)", type=Path, default="-")

    parser.add_argument("-o", "--output-file", dest="output_file_path",
                        help="The KGTK output file. (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--reified-file", dest="reified_file_path",
                              help="A KGTK output file that will contain only the reified values. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--unreified-file", dest="unreified_file_path",
                              help="A KGTK output file that will contain only the unreified values. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--uninvolved-file", dest="uninvolved_file_path",
                              help="A KGTK output file that will contain only the uninvolved input records. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    KgtkUnreifyValues.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=False, expert=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("--input-files %s" % " ".join([str(path) for  path in input_file_paths]), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        if args.reified_file_path is not None:
            print("--reified-file=%s" % str(args.reified_file_path), file=error_file, flush=True)
        if args.unreified_file_path is not None:
            print("--unreified-file=%s" % str(args.unreified_file_path), file=error_file, flush=True)
        if args.uninvolved_file_path is not None:
            print("--uninvolved-file=%s" % str(args.uninvolved_file_path), file=error_file, flush=True)

        if args.output_format is not None:
            print("--output-format=%s" % args.output_format, file=error_file, flush=True)

        if args.trigger_label_value is not None:
            print("--trigger-label=%s" % args.trigger_label_value, file=error_file, flush=True)
        if args.trigger_node2_value is not None:
            print("--trigger-node2=%s" % args.trigger_node2_value, file=error_file, flush=True)
        if args.value_label_value is not None:
            print("--value-label=%s" % args.value_label_value, file=error_file, flush=True)
        if args.old_label_value is not None:
            print("--old-label=%s" % args.old_label_value, file=error_file, flush=True)
        if args.new_label_value is not None:
            print("--new-label=%s" % args.new_label_value, file=error_file, flush=True)

        print("--allow-multiple-values=%s" % str(args.allow_multiple_values), file=error_file, flush=True)
        print("--allow-extra-columns=%s" % str(args.allow_extra_columns), file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kuv: KgtkUnreifyValues = KgtkUnreifyValues(
        input_file_path=args.input_file_path,
        output_file_path=args.output_file_path,
        reified_file_path=args.reified_file_path,
        unreified_file_path=args.unreified_file_path,
        uninvolved_file_path=args.uninvolved_file_path,

        trigger_label_value=args.trigger_label_value,
        trigger_node2_value=args.trigger_node2_value,
        value_label_value=args.value_label_value,
        old_label_value=args.old_label_value,
        new_label_value=args.new_label_value,

        allow_multiple_values=args.allow_multiple_values,
        allow_extra_columns=args.allow_extra_columns,

        reader_options=reader_options,
        value_options=value_options,
        output_format=args.output_format,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose,
    )

    kuv.process()
    
if __name__ == "__main__":
    main()



# text embedding lexicalization.
#
# TODO: Provide seperate KgtkReader options (with fallback) for
# the entity label files, and the main input files.
#
# Approach:
# The input file is assumed to be sorted by node1 value (lowest to highest).
# Read each group of records with the same node1 value and construct
# an English sentence that describes the entity and its properties.
#
# Processing an entity may need English labels for the relationships (label column)
# and other entities (node2 column) to which it is related. The lable values must
# be loaded into memory first.  This can be accomplished by splitting the English labels
# into a separate file and passing it to  --entity-label-file, or by reading
# the input file twice (once as --input-file and once as --entity-label-file, both
# of which must be specified on the command line).
#
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

DEFAULT_LABEL_PROPERTIES: typing.List[str] = [ "label" ]
DEFAULT_DESCRIPTION_PROPERTIES: typing.List[str] = [ "description" ]
DEFAULT_ISA_PROPERTIES: typing.List[str] = [ "P21", "P31", "P39", "P106", "P279" ]
DEFAULT_HAS_PROPERTIES: typing.List[str] = [ ]
DEFAULT_PROPERTY_VALUES: typing.List[str] = [ "P17" ]
DEFAULT_SENTENCE_LABEL: str = "sentence"

OUTPUT_COLUMNS: typing.List[str] = [ "node1", "label", "node2" ]

def parser():
    return {
        'help': """Produce sentences from a KGTK file."""
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace ):
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file()
    parser.add_input_file(who="The entity label file(s)",
                          dest="entity_label_files",
                          options=['--entity-label-file'],
                          metavar="ENTITY_LABEL_FILE",
                          optional=True,
                          allow_list=True,
                          default_stdin=False)
    parser.add_output_file()

    parser.add_argument("--label-properties", dest="label_properties", nargs="*",
                        help="The label properties. (default=%s)" % repr(DEFAULT_LABEL_PROPERTIES))

    parser.add_argument("--description-properties", dest="description_properties", nargs="*",
                        help="The description properties. (default=%s)" % repr(DEFAULT_DESCRIPTION_PROPERTIES))

    parser.add_argument("--isa-properties", dest="isa_properties", nargs="*",
                        help="The isa properties. (default=%s)" % repr(DEFAULT_ISA_PROPERTIES))

    parser.add_argument("--has-properties", dest="has_properties", nargs="*",
                        help="The has properties. (default=%s)" % repr(DEFAULT_HAS_PROPERTIES))

    parser.add_argument("--property-values", dest="property_values", nargs="*",
                        help="The property values. (default=%s)" % repr(DEFAULT_PROPERTY_VALUES))

    parser.add_argument('--sentence-label', action='store', type=str, dest='sentence_label', default=DEFAULT_SENTENCE_LABEL,
                        help="The relationship to write in the output file. (default=%(default)s)")

    parser.add_argument("--explain", dest="explain", metavar="True|False",
                        help="When true, include an explanation column that tells how the sentence was constructed. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument("--presorted", dest="presorted", metavar="True|False",
                        help="When true, the input file is presorted on node1. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument("--add-entity-labels-from-input", dest="add_entity_labels_from_input", metavar="True|False",
                        help="When true, extract entity labels from the unsorted input file. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=False)
    KgtkValueOptions.add_arguments(parser, expert=False)


def run(input_file: KGTKFiles,
        entity_label_files: KGTKFiles,
        output_file: KGTKFiles,

        label_properties: typing.Optional[typing.List[str]],
        description_properties: typing.Optional[typing.List[str]],
        isa_properties: typing.Optional[typing.List[str]],
        has_properties: typing.Optional[typing.List[str]],
        property_values: typing.Optional[typing.List[str]],
        sentence_label: str,
        explain: bool,
        presorted: bool,
        add_entity_labels_from_input: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
        
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException

    from kgtk.gt.lexicalize_utils import Lexicalize

    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    entity_label_kgtk_files: typing.List[Path] = KGTKArgumentParser.get_input_file_list(entity_label_files,
                                                                                        who="The entity label file(s)",
                                                                                        default_stdin=False)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    if label_properties is None:
        label_properties = DEFAULT_LABEL_PROPERTIES

    if description_properties is None:
        description_properties = DEFAULT_DESCRIPTION_PROPERTIES

    if isa_properties is None:
        isa_properties = DEFAULT_ISA_PROPERTIES

    if has_properties is None:
        has_properties = DEFAULT_HAS_PROPERTIES

    if property_values is None:
        property_values = DEFAULT_PROPERTY_VALUES

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file, flush=True)
        if len(entity_label_kgtk_files) > 0:
            print("--entity-label-files %s" % " ".join([str(f) for f in entity_label_kgtk_files]), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)

        if len(label_properties) > 0:
            print("--label-properties %s" % " ".join(label_properties), file=error_file, flush=True)

        if len(description_properties) > 0:
            print("--description-properties %s" % " ".join(description_properties), file=error_file, flush=True)

        if len(isa_properties) > 0:
            print("--isa-properties %s" % " ".join(isa_properties), file=error_file, flush=True)

        if len(has_properties) > 0:
            print("--has-properties %s" % " ".join(has_properties), file=error_file, flush=True)

        if len(property_values) > 0:
            print("--property-values %s" % " ".join(property_values), file=error_file, flush=True)

        print("--sentence-label=%s" % str(sentence_label), file=error_file, flush=True)
        print("--explain=%s" % str(explain), file=error_file, flush=True)
        print("--presorted=%s" % str(presorted), file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)


    lexer: Lexicalize = Lexicalize(label_properties,
                                   description_properties,
                                   isa_properties,
                                   has_properties,
                                   property_values,
                                   sentence_label,
                                   explain=explain,
                                   error_file=error_file,
                                   verbose=verbose,
                                   very_verbose=very_verbose)
    if len(entity_label_kgtk_files) > 0:
        lexer.load_entity_label_files(entity_label_kgtk_files,
                                      error_file,
                                      reader_options,
                                      value_options,
                                      label_properties=label_properties,
                                      verbose=verbose)
        
    kr: typing.Optional[KgtkReader] = None
    kw: typing.Optional[KgtkWriter] = None

    try:
        if verbose:
            print("Opening the input file %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr = KgtkReader.open(input_kgtk_file,
                             options=reader_options,
                             value_options = value_options,
                             error_file=error_file,
                             verbose=verbose,
                             very_verbose=very_verbose,
                             )

        if kr.node1_column_idx < 0:
            raise KGTKException("Missing column: node1 or alias")
        if kr.label_column_idx < 0:
            raise KGTKException("Missing column: label or alias")
        if kr.node2_column_idx < 0:
            raise KGTKException("Missing column: node2 or alias")

        if verbose:
            print("node1 column index = {}".format(kr.node1_column_idx),  file=error_file, flush=True)
            print("label column index = {}".format(kr.label_column_idx),  file=error_file, flush=True)
            print("node2 column index = {}".format(kr.node2_column_idx),  file=error_file, flush=True)

        output_columns: typing.List[str] = OUTPUT_COLUMNS.copy()
        if explain:
            output_columns.append("explaination")
            if verbose:
                print("Including an explaination column in the output.", file=error_file, flush=True)

        if verbose:
            print("Opening the output file %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw = KgtkWriter.open(output_columns,
                             output_kgtk_file,
                             require_all_columns=True,
                             prohibit_extra_columns=True,
                             fill_missing_columns=False,
                             gzip_in_parallel=False,
                             verbose=verbose,
                             very_verbose=very_verbose,
                             )

        if presorted:
            lexer.process_presorted_input(kr, kw)
        else:
            lexer.process_unsorted_input(kr, kw, add_entity_labels=add_entity_labels_from_input)

        return 0

    except Exception as e:
        raise KGTKException(str(e))

    finally:
        if kw is not None:
            kw.close()
            
        if kr is not None:
            kr.close()



import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry
from kgtk.exceptions import KGTKException

class TestImportWordNet(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir=tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_import(self):
        cli_entry("kgtk", "import_wordnet", "-o", f'{self.temp_dir}/wordnet.tsv')

        df = pd.read_csv(f'{self.temp_dir}/wordnet.tsv', sep='\t')

        self.assertEqual(len(df.columns), 9)

        relations = list(df['relation'].unique())

        for r in ['/r/IsA', '/r/PartOf', '/r/MadeOf']:
            self.assertTrue(r in relations)




"""
Generate graph embedding based on Pytorch BigGraph library  

"""

from argparse import Namespace
from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
import attr,sys,typing
from kgtk.kgtkformat import KgtkFormat
from pathlib import Path
import sys

import os
os.environ['KMP_DUPLICATE_LIB_OK']='True' # remove the Issue: Initializing libiomp5.dylib, but found libiomp5.dylib already initialized.


@attr.s(slots=True, frozen=False)
class KgtkCreateTmpTsv(KgtkFormat):
    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process(self):
        # Open the input file.
        if self.verbose:
            print("Opening the input file: %s" % str(self.input_file_path), file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,                          
        )
        
        if self.verbose:
            print("Opening the output file: %s" % str(self.output_file_path), file=self.error_file, flush=True)

        # Open the output file.
        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)
        # here kw has one line already where PBG doesn't need it,

        input_line_count: int = 0
        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        # node1 relation node2
        node1_index= kr.get_node1_column_index()
        node2_index = kr.get_node2_column_index()
        ##relation_index = kr.get_id_column_index('relation')#
        relation_index = kr.get_label_column_index()
      
        row: typing.List[str]
        # delete header
        # kw.file_out.seek(0)         # set the cursor to the top of the file
        # kw.file_out.truncate()      # truncate following part == delete first line
        # print(kw.file_out.tell())

        for row in kr:
            input_line_count += 1
            kw.write([row[node1_index],row[relation_index],row[node2_index]]) 
           
        if self.verbose:
            print("Processed %d records." % (input_line_count), file=self.error_file, flush=True)
    
        kw.close()

def get_config(**kwargs): 
    """
    configurations for graph embedding
    """
    # import modules locally
    from pathlib import Path

    output_folder = kwargs['temporary_directory'] / 'output'
    entity_path = str(output_folder)
    edge_paths = [ str((output_folder / 'edges_partitioned')) ]
    checkpoint_path = str((output_folder / 'model'))
       
    config = dict(
        # I/O data
        entity_path = entity_path,
        edge_paths = edge_paths,
        checkpoint_path = checkpoint_path,
        # Graph structure
        entities= {"all": {"num_partitions": 1 }}  ,
        relations=[  # relation template setting
        {
            "name": "all_edges",
            "lhs": "all",
            "rhs": "all",
            "operator": kwargs['operator'],
        }
        ],
        dynamic_relations=kwargs['dynamic_relaitons'],
        # Scoring model
        dimension=kwargs['dimension_num'],
        global_emb=kwargs['global_emb'],
        comparator=kwargs['comparator'],
        # Training
        init_scale = kwargs['init_scale'],
        bias=kwargs['bias'],
        num_epochs=kwargs['num_epochs'],
        loss_fn=kwargs['loss_fn'],
        lr=kwargs['learning_rate'],
        # Evaluation during training
        eval_fraction=kwargs['eval_fraction'],  # to reproduce results, we need to use all training data 
    )

    return config

def parser():
    return {
        'help': 'graph embedding functionality',
        'description': 'Generate graph embedding in kgtk tsv format, here we use PytorchBigGraph as \
        low-level implementation '
    }

def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    ### IO 
    parser.add_argument(      "-i", "--input-file", dest="input_file_path",
                              help="The KGTK input file. (default=%(default)s)", type=Path, default="-")
    parser.add_argument(      "-o", "--output-file", dest="output_file_path",
                              help="The KGTK output file. (default=%(default)s).", type=Path, default="-")
    parser.add_argument(     '-l',"--log", dest="log_file_path",
                              help="Setting the log path [Default: None]",
                              type=Path,default=None, metavar="")
    parser.add_argument(     '-T','--temporary_directory', dest='temporary_directory',
                             help="Sepecify the directory location to store temporary file",
                             type=Path,default=Path('/tmp/'), metavar='')
    parser.add_argument(     '-ot','--output_format', dest='output_format',
                             help="Outputformat for embeddings [Default: w2v] Choice: kgtk | w2v | glove",
                             default='w2v', metavar='')
    parser.add_argument(     '-r','--retain_temporary_data', dest='retain_temporary_data',
                            help="When opearte graph, some tempory files will be generated, set True to retain these files ",
                            type=bool,default=True, metavar='True|False')
    ### Training parameters
    parser.add_argument(     '-d',"--dimension", dest="dimension_num",
                              help="Dimension of the real space the embedding live in [Default: 100]",
                              type=int,default=100, metavar="")
    parser.add_argument(     '-s',"--init_scale", dest="init_scale",
                              help="Generating the initial embedding with this standard deviation [Default: 0.001]" +
                              "If no initial embeddings are provided, they are generated by sampling each dimension" +
                              "from a centered normal distribution having this standard deviation.",
                              type=float,default=0.001, metavar="")
    parser.add_argument(     '-c','--comparator', dest='comparator',
                             help="How the embeddings of the two sides of an edge (after having already "+
                             "undergone some processing) are compared to each other to produce a score[Default: dot]," +
                             "Choice: dot|cos|l2|squared_l2",
                             default='dot',choices=['dot','cos','l2','squared_l2'],metavar='dot|cos|l2|squared_l2')
    parser.add_argument(     '-op','--operator', dest='operator',
                             help="The transformation to apply to the embedding of one of the sides of the edge " +
                             "(typically the right-hand one) before comparing it with the other one. It reflects which model that embedding uses. " +
                             "[Default:ComplEx]",
                              #default will be setting to complex_diagonal later
                             default='ComplEx',metavar='RESCAL|DistMult|ComplEx|TransE')
    parser.add_argument(     '-e','--num_epochs', dest='num_epochs',
                             help="The number of times the training loop iterates over all the edges.[Default:100]",
                             type=int,default=100, metavar='')    
    parser.add_argument(      '-b','--bias', dest='bias',
                             help="Whether use the bias choice [Default: False],If enabled, withhold the first "+
                             "dimension of the embeddings from the comparator and instead use it as a bias, adding "+
                             "back to the score. Makes sense for logistic and softmax loss functions. ",
                             type=bool,default=False, metavar='True|False')
    parser.add_argument(     '-w','--workers', dest='workers',
                             help="The number of worker processes for training. If not given, set to CPU count.",
                             type=int,default=None, metavar='')
    parser.add_argument(     '-bs','--batch_size', dest='batch_size',
                             help="The number of edges per batch.[Default:1000]",
                             type=int,default=1000, metavar='')
    parser.add_argument(     '-lf','--loss_fn', dest='loss_fn',
                             help="How the scores of positive edges and their corresponding negatives "+
                             "are evaluated.[Default: ranking], Choice: ranking|logistic|softmax",
                             # default will be setting to ranking later
                             default=None,choices=['ranking','logistic','softmax',None],metavar='ranking|logistic|softmax')
    parser.add_argument(     '-lr','--learning_rate', dest='learning_rate',
                             help="The learning rate for the optimizer.[Default: 0.1]",
                             # default will be setting to 0.1 later
                             type=float,default=None,metavar='')
    parser.add_argument(     '-ef','--eval_fraction', dest='eval_fraction',
                             help="The fraction of edges withheld from training and used to track evaluation "+
                            "metrics during training. [Defalut:0.0 training all edges ]",
                             type=float,default=0.0,metavar='')
    parser.add_argument(     '-dr','--dynamic_relaitons', dest='dynamic_relaitons',
                             help="Whether use dynamic relations (when graphs with a "+ 
	                         "large number of relations) [Default: True]",
                             type=bool, default=True, metavar='True|False')             
    parser.add_argument(     '-ge','--global_emb', dest='global_emb',
                             help="Whether use global embedding, if enabled, add to each embedding a vector that is common "
                             "to all the entities of a certain type. This vector is learned during training.[Default: False] ",
                             type=bool,default=False, metavar='True|False')
    ### kgtk format
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=True)
    KgtkValueOptions.add_arguments(parser)


def config_preprocess(raw_config):
    ## Setting learning rate and loss function for different algo if user doesn't assign them
    '''
    loss_fn: ranking
    learning_rate: 0.01
    operator:complex_diagonal
    '''

    algorithm_operator = {"complex":"complex_diagonal", 
                          "distmult": "diagonal", 
                          "rescal":"linear",
                          "transe":"translation"}
    try:
        algorithm = algorithm_operator[raw_config['relations'][0]['operator'].lower()]
        raw_config['relations'][0]['operator'] = algorithm
    except:
        print('Plase use valid operator! choices: RESCAL|DistMult|ComplEx|TransE')
        import sys
        sys.exit()

    loss_fn = raw_config['loss_fn']
    learning_rate = raw_config['lr']
    if algorithm and loss_fn and learning_rate:
        return 

    if  not algorithm:  # use doesn't enter anything
        raw_config['relations'][0]['operator'] = 'complex_diagonal'
        if not loss_fn:
            raw_config['loss_fn'] = 'logistic'
        if not learning_rate:
            raw_config['lr'] = 0.1
    if algorithm: # give the algorithm 
        if algorithm == 'complex_diagonal': # ComplEx
            if not loss_fn:
                raw_config['loss_fn'] = 'logistic'
            if not learning_rate:
                raw_config['lr'] = 0.1
        elif algorithm == 'translation': # TransE
            if not loss_fn:
                raw_config['loss_fn'] = 'logistic'
            if not learning_rate:
                raw_config['lr'] = 0.1
        elif algorithm == 'diagonal': # DistMult
            if not loss_fn:
                raw_config['loss_fn'] = 'ranking'
            if not learning_rate:
                raw_config['lr'] = 0.01
        else: # RESCAL
            if not loss_fn:
                raw_config['loss_fn'] = 'ranking'
            if not learning_rate:
                raw_config['lr'] = 0.01

    processed_config = raw_config
    return processed_config

# convert wv format to kgtk format ..
def generate_kgtk_output(entities_output,output_kgtk_file,verbose,very_verbose):

    # Open the output file.
    kw: KgtkWriter = KgtkWriter.open(#kr.column_names,
                                    ['id','node1','node2','relation'], # in order to obey the kgtk rules
                                    output_kgtk_file,
                                    #mode=KgtkWriter.Mode[kr.mode.name],
                                    mode = KgtkWriter.Mode.AUTO,
                                    require_all_columns=False,
                                    prohibit_extra_columns=False,
                                    fill_missing_columns=False,
                                    gzip_in_parallel=False,
                                    verbose=verbose,
                                    very_verbose=very_verbose)

    input_line_count: int = 0
    if verbose:
        logging.info("Processing the input records.", file=self.error_file, flush=True)

    #delete header
    kw.file_out.seek(0)         # set the cursor to the top of the file
    kw.file_out.truncate()      # truncate following part == delete first line

    MODULE_NAME = 'graph_embeddings' # __name__.split('.')[-1] 
    with open(entities_output) as wv_file:
        for line in wv_file:
            line = line.replace('\n','')   #remove  \n
            entity_name = line.split('\t')[0]
            entity_vev = ','.join(line.split('\t')[1:])
            input_line_count += 1
            kw.write([entity_name,MODULE_NAME,entity_vev]) 
             
    if verbose:
        logging.info("Processed %d records." % (input_line_count), file=self.error_file, flush=True)

    kw.close()

def generate_w2v_output(entities_output,output_kgtk_file,kwargs):
    fout = open(output_kgtk_file,'w')
    fin = open(entities_output)
    entity_num = len(fin.readlines())
    fin.close()
    fout.write(str(entity_num) + ' ' + str(kwargs['dimension_num']) + '\n')
    with open(entities_output) as fin:
        for line in fin:
            embedding = ' '.join(line.split('\t'))
            fout.write(embedding)
    fout.close()

def run(verbose: bool = False,
        very_verbose: bool = False,
        **kwargs):
    """
    **kwargs stores all parameters providing by user
    """
    # print(kwargs)

    # import modules locally
    import sys
    import typing
    import os
    import logging
    from pathlib import Path
    import json,os,h5py,gzip,torch,shutil
    from torchbiggraph.config import parse_config
    # copy  missing file under kgtk/graph_embeddings
    from kgtk.graph_embeddings.importers import TSVEdgelistReader,convert_input_data  
    from torchbiggraph.train import train
    from torchbiggraph.util import SubprocessInitializer, setup_logging
    from kgtk.graph_embeddings.export_to_tsv import make_tsv
    # from torchbiggraph.converters.export_to_tsv import make_tsv

    try:
        # store the data into log file, then the console will not output anything
        if kwargs['log_file_path'] != None: 
            log_file_path = kwargs['log_file_path']
            logging.basicConfig(format='%(asctime)s - %(filename)s[line:%(lineno)d] \
            - %(levelname)s: %(message)s',
                    level=logging.DEBUG,
                    filename=str(log_file_path),
                    filemode='w')
            print(f'In Processing, Please go to {kwargs["log_file_path"]} to check details')

        input_kgtk_file: Path = kwargs['input_file_path']
        tmp_folder = kwargs['temporary_directory']
        tmp_tsv_path:Path = tmp_folder / f'tmp_{input_kgtk_file.name}'
        # tmp_tsv_path:Path = input_kgtk_file.parent/f'tmp_{input_kgtk_file.name}'

        #  make sure the tmp folder exists, otherwise it will raise an exception
        if not os.path.exists(tmp_folder):
            os.makedirs(tmp_folder)

        output_kgtk_file = kwargs['output_file_path']
        try:   #if output_kgtk_file is not empty, delete it
            output_kgtk_file.unlink() 
        except: pass # didn't find, then let it go

        # *********************************************
        # 0. PREPARE PBG TSV FILE
        # *********************************************
        reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)
        error_file: typing.TextIO = sys.stdout if kwargs.get("errors_to_stdout") else sys.stderr
        kct: KgtkCopyTemplate = KgtkCreateTmpTsv(
            input_file_path=input_kgtk_file,
            output_file_path=tmp_tsv_path,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        # prepare the graph file
        # create a tmp tsv file for PBG embedding
        
        logging.info('Generate the valid tsv format for embedding ...')
        kct.process()
        logging.info('Embedding file is ready...')
        
        # *********************************************
        # 1. DEFINE CONFIG  
        # *********************************************
        raw_config = get_config(**kwargs)

        ## setting corresponding learning rate and loss function for different algorthim
        processed_config = config_preprocess(raw_config)
        
        # temporry output folder 
        tmp_output_folder = Path(processed_config['entity_path'])

        # before moving, need to check whether the tmp folder is not empty in case of bug
        try: #if temporry output folder is alrady existing then delete it
            shutil.rmtree(tmp_output_folder)
        except: pass # didn't find, then let it go

        # **************************************************
        # 2. TRANSFORM GRAPH TO A BIGGRAPH-FRIENDLY FORMAT
        # **************************************************
        setup_logging()
        config = parse_config(processed_config)
        subprocess_init = SubprocessInitializer()
        input_edge_paths = [tmp_tsv_path] 

        convert_input_data(
            config.entities,
            config.relations,
            config.entity_path,
            config.edge_paths,
            input_edge_paths,
            TSVEdgelistReader(lhs_col=0, rel_col=1, rhs_col=2),
            dynamic_relations=config.dynamic_relations,
        )

        # ************************************************
        # 3. TRAIN THE EMBEDDINGS
        #*************************************************
        train(config, subprocess_init=subprocess_init)

        # ************************************************
        # 4. GENERATE THE OUTPUT
        # ************************************************
        # entities_output = output_kgtk_file
        entities_output = tmp_output_folder/'entities_output.tsv'
        relation_types_output = tmp_output_folder / 'relation_types_tf.tsv'

        with open(entities_output, "xt") as entities_tf, open(
            relation_types_output, "xt"
        ) as relation_types_tf:
            make_tsv(config, entities_tf, relation_types_tf)

        # output  correct format for embeddings
        if kwargs['output_format'] == 'glove': # glove format output 
            shutil.copyfile(entities_output,output_kgtk_file)
        elif kwargs['output_format'] == 'w2v': # w2v format output
            generate_w2v_output(entities_output,output_kgtk_file,kwargs)

        else: # write to the kgtk output format tsv 
            generate_kgtk_output(entities_output,output_kgtk_file,verbose,very_verbose)

        logging.info(f'Embeddings has been generated in {output_kgtk_file}.')

        # ************************************************
        # 5. Garbage collection  
        # ************************************************
        if kwargs['retain_temporary_data'] == False:
            shutil.rmtree(kwargs['temporary_directory'])
            # tmp_tsv_path.unlink() # delete temporay tsv file
            # shutil.rmtree(tmp_output_folder) # deleter temporay output folder   

        if kwargs["log_file_path"] != None:
            print('Processed Finished.')
            logging.info(f"Process Finished.\nOutput has been saved in {kwargs['output_file_path']}")
        else:
            print(f"Process Finished.\nOutput has been saved in {kwargs['output_file_path']}")

    except Exception as e:
        raise KGTKException(str(e))
            


"""
Unreify RDF statements in KGTK files.

"""
from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.unreify.kgtksortbuffer import KgtkSortBuffer
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=False)
class KgtkUnreifyRdfStatements(KgtkFormat):
    DEFAULT_TRIGGER_LABEL_VALUE: str = "rdf:type"
    DEFAULT_TRIGGER_NODE2_VALUE: str = "rdf:Statement"
    DEFAULT_RDF_OBJECT_LABEL_VALUE: str = "rdf:object"
    DEFAULT_RDF_PREDICATE_LABEL_VALUE: str = "rdf:predicate"
    DEFAULT_RDF_SUBJECT_LABEL_VALUE: str = "rdf:subject"
    DEFAULT_ALLOW_MULTIPLE_SUBJECTS: bool = True
    DEFAULT_ALLOW_MULTIPLE_PREDICATES: bool = True
    DEFAULT_ALLOW_MULTIPLE_OBJECTS: bool = True

    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reified_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    unreified_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    uninvolved_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    trigger_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_TRIGGER_LABEL_VALUE)
    trigger_node2_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_TRIGGER_NODE2_VALUE)
    rdf_object_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_RDF_OBJECT_LABEL_VALUE)
    rdf_predicate_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_RDF_PREDICATE_LABEL_VALUE)
    rdf_subject_label_value: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_RDF_SUBJECT_LABEL_VALUE)

    allow_multiple_subjects: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_MULTIPLE_SUBJECTS)
    allow_multiple_predicates: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_MULTIPLE_PREDICATES)
    allow_multiple_objects: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_MULTIPLE_OBJECTS)

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    output_format: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None) # TODO: use an enum

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # Working variables:
    output_line_count: int = attr.ib(default=0)

    def process(self):
        # Open the input file.
        if self.verbose:
            print("Opening the input file: %s" % str(self.input_file_path), file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          mode=KgtkReaderMode.EDGE, # Must be an edge file.
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        output_column_names: typing.List[str] = kr.column_names.copy()

        node1_column_idx: int = kr.node1_column_idx
        node1_column_name: str = output_column_names[node1_column_idx]

        label_column_idx: int = kr.label_column_idx
        label_column_name: str = output_column_names[label_column_idx]

        node2_column_idx: int = kr.node2_column_idx
        node2_column_name: str = output_column_names[node2_column_idx]

        # Adding an ID column?
        new_id_column: bool = False
        id_column_idx: int = kr.id_column_idx
        if id_column_idx < 0:
            new_id_column = True
            id_column_idx = len(output_column_names)
            output_column_names.append(KgtkFormat.ID)
        id_column_name: str = output_column_names[id_column_idx]

        if self.verbose:
            print("Opening the output file: %s" % str(self.output_file_path), file=self.error_file, flush=True)
        # Open the output file.
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         output_format=self.output_format,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        reifiedw: typing.Optional[KgtkWriter] = None
        if self.reified_file_path is not None:
            if self.verbose:
                print("Opening the reified RDF statements output file: %s" % str(self.reified_file_path), file=self.error_file, flush=True)
            reifiedw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                                   self.reified_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=True,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=False,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        unreifiedw: typing.Optional[KgtkWriter] = None
        if self.unreified_file_path is not None:
            if self.verbose:
                print("Opening the unreified RDF statements output file: %s" % str(self.unreified_file_path), file=self.error_file, flush=True)
            unreifiedw: KgtkWriter = KgtkWriter.open(output_column_names,
                                                   self.unreified_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=True,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=False,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        uninvolvedw: typing.Optional[KgtkWriter] = None
        if self.uninvolved_file_path is not None:
            if self.verbose:
                print("Opening the uninvolved records output file: %s" % str(self.uninvolved_file_path), file=self.error_file, flush=True)
            uninvolvedw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                                   self.uninvolved_file_path,
                                                   mode=KgtkWriter.Mode[kr.mode.name],
                                                   output_format=self.output_format,
                                                   require_all_columns=True,
                                                   prohibit_extra_columns=True,
                                                   fill_missing_columns=False,
                                                   gzip_in_parallel=False,
                                                   verbose=self.verbose,
                                                   very_verbose=self.very_verbose)

        if self.verbose:
            print("Reading and grouping the input records.", file=self.error_file, flush=True)
        ksb: KgtkSortBuffer = KgtkSortBuffer.readall(kr, grouped=True, keygen=KgtkSortBuffer.node1_keygen)

        input_group_count: int = 0
        input_line_count: int = 0
        self.output_line_count = 0
        unreification_count: int = 0

        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        node1_group: typing.List[typing.List[str]]
        for node1_group in ksb.groupiterate():
            input_group_count += 1

            saw_error: bool = False
            saw_trigger: bool = False
            node1_value: typing.Optional[str] = None
            rdf_object_values: typing.Set[str] = set()
            rdf_predicate_values: typing.Set[str] = set()
            rdf_subject_values: typing.Set[str] = set()
            
            potential_edge_attributes: typing.List[typing.List[str]] = [ ]

            row: typing.List[str]
            for row in node1_group:
                input_line_count += 1
                if node1_value is None:
                    node1_value = row[node1_column_idx]
                label: str = row[label_column_idx]
                node2: str = row[node2_column_idx]

                if label == self.trigger_label_value and node2 == self.trigger_node2_value:
                    if saw_trigger:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Duplicate trigger in input group %d (%s)" % (input_group_count, node1_value), file=self.error_file, flush=True)
                    saw_trigger = True
                elif label == self.rdf_object_label_value:
                    if len(rdf_object_values) > 0 and node2 not in rdf_object_values and not self.allow_multple_objects:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Multiple rdf objects in input group %d (%s)" % (input_group_count, node1_value), file=self.error_file, flush=True)
                        saw_error = True # until we implement the cartesan product
                    rdf_object_values.add(node2)
                elif label == self.rdf_predicate_label_value:
                    if len(rdf_predicate_values) > 0  and node2 not in rdf_predicate_values and not self.allow_multiple_predicates:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Multiple rdf predicates in input group %d (%s)" % (input_group_count, node1_value), file=self.error_file, flush=True)
                        saw_error = True # until we implement the cartesan product
                    rdf_predicate_values.add(node2)
                elif label == self.rdf_subject_label_value:
                    if len(rdf_subject_values) > 0 and node2 not in rdf_subject_values and not self.allow_multple_subjects:
                        # TODO: Shout louder.
                        if self.verbose:
                            print("Warning: Multiple rdf subjects in input group %d (%s)" % (input_group_count, node1_value), file=self.error_file, flush=True)
                        saw_error = True # until we implement the cartesan product
                    rdf_subject_values.add(node2)
                else:
                    potential_edge_attributes.append(row)
                    
            rdf_product: int = len(rdf_object_values) * len(rdf_predicate_values) * len(rdf_subject_values)

            if saw_trigger and \
               node1_value is not None and \
               rdf_product > 0 and \
               not saw_error:
                # Unreification was triggered.
                unreification_count += 1

                self.write_new_edge_or_edges(kw,
                                             reifiedw,
                                             unreifiedw,
                                             potential_edge_attributes,
                                             node1_group,
                                             rdf_product,
                                             list(rdf_subject_values),
                                             list(rdf_predicate_values),
                                             list(rdf_object_values),
                                             node1_value,
                                             label_column_idx,
                                             node2_column_idx,
                                             node1_column_name,
                                             label_column_name,
                                             node2_column_name,
                                             id_column_name,
                )
                
            else:
                # Unreification was not triggered.  Pass this group of rows
                # through unchanged, except for possibly appending an ID
                # column.
                self.pass_group_through(kw, uninvolvedw, node1_group, new_id_column)

        if self.verbose:
            print("Processed %d records in %d groups." % (input_line_count, input_group_count), file=self.error_file, flush=True)
            print("Unreified %d groups." % unreification_count, file=self.error_file, flush=True)
            print("Wrote %d output records" % self.output_line_count, file=self.error_file, flush=True)

        
        kw.close()
        if reifiedw is not None:
            reifiedw.close()
        if unreifiedw is not None:
            unreifiedw.close()
        if uninvolvedw is not None:
            uninvolvedw.close()

    def write_new_edge_or_edges(self,
                                kw: KgtkWriter,
                                reifiedw: typing.Optional[KgtkWriter],
                                unreifiedw: typing.Optional[KgtkWriter],
                                potential_edge_attributes: typing.List[typing.List[str]],
                                node1_group: typing.List[typing.List[str]],
                                rdf_product: int,
                                rdf_subject_values: typing.List[str],
                                rdf_predicate_values: typing.List[str],
                                rdf_object_values: typing.List[str],
                                edge_id: str,
                                label_column_idx: int,
                                node2_column_idx: int,
                                node1_column_name: str,
                                label_column_name: str,
                                node2_column_name: str,
                                id_column_name: str,
    ):
        if reifiedw is not None:
            for row in node1_group:
                reifiedw.write(row)
        if rdf_product == 1:
            # Generate the new edge:
            self.write_new_edge(kw,
                                unreifiedw,
                                potential_edge_attributes,
                                edge_id,
                                rdf_subject_values[0],
                                rdf_predicate_values[0],
                                rdf_object_values[0],
                                label_column_idx,
                                node2_column_idx,
                                node1_column_name,
                                label_column_name,
                                node2_column_name,
                                id_column_name,
            )
        else:
            width: int = self.get_width(rdf_product)
            edge_number: int = 0
            rdf_subject_value: str
            for rdf_subject_value in sorted(rdf_subject_values):
                rdf_predicate_value: str
                for rdf_predicate_value in sorted(rdf_predicate_values):
                    rdf_object_value: str
                    for rdf_object_value in sorted(rdf_object_values):
                        edge_number += 1
                        new_edge_id: str = self.make_new_id(edge_id, edge_number, width)
                        self.write_new_edge(kw,
                                            unreifiedw,
                                            potential_edge_attributes,
                                            new_edge_id,
                                            rdf_subject_value,
                                            rdf_predicate_value,
                                            rdf_object_value,
                                            label_column_idx,
                                            node2_column_idx,
                                            node1_column_name,
                                            label_column_name,
                                            node2_column_name,
                                            id_column_name,
            )


    def make_new_id(self, edge_id: str, count: int, width: int)->str:
        # Generate a new ID that will sort after the new edge.
        # What if the existing ID is not a symbol or a string?
        #
        # TODO: Handle these cases.
        new_id: str
        if edge_id.startswith(KgtkFormat.STRING_SIGIL) and edge_id.endswith(KgtkFormat.STRING_SIGIL):
            new_id = edge_id[:-1] + "-" + str(count).zfill(width) + KgtkFormat.STRING_SIGIL
        else:
            new_id = edge_id + "-" + str(count).zfill(width)
        return new_id

    def get_width(self, max_count: int)->int:
        return len(str(max_count).strip())
        
    def write_new_edge(self,
                       kw: KgtkWriter,
                       unreifiedw: typing.Optional[KgtkWriter],
                       potential_edge_attributes: typing.List[typing.List[str]],
                       edge_id: str,
                       rdf_subject_value: str,
                       rdf_predicate_value: str,
                       rdf_object_value: str,
                       label_column_idx: int,
                       node2_column_idx: int,
                       node1_column_name: str,
                       label_column_name: str,
                       node2_column_name: str,
                       id_column_name: str,
    ):
        kw.writemap({
            node1_column_name: rdf_subject_value,
            label_column_name: rdf_predicate_value,
            node2_column_name: rdf_object_value,
            id_column_name: edge_id,
        })
        self.output_line_count += 1
        
        if unreifiedw is not None:
            unreifiedw.writemap({
                node1_column_name: rdf_subject_value,
                label_column_name: rdf_predicate_value,
                node2_column_name: rdf_object_value,
                id_column_name: edge_id,
            })

        self.write_edge_attributes(kw,
                                   unreifiedw,
                                   potential_edge_attributes,
                                   edge_id,
                                   label_column_idx,
                                   node2_column_idx,
                                   node1_column_name,
                                   label_column_name,
                                   node2_column_name,
                                   id_column_name,
)

    def write_edge_attributes(self,
                              kw: KgtkWriter,
                              unreifiedw: typing.Optional[KgtkWriter],
                              potential_edge_attributes: typing.List[typing.List[str]],
                              edge_id: str,
                              label_column_idx: int,
                              node2_column_idx: int,
                              node1_column_name: str,
                              label_column_name: str,
                              node2_column_name: str,
                              id_column_name: str,
    ):
        width: int = self.get_width(len(potential_edge_attributes))
        attribute_number: int = 0
        edge_row: typing.List[str]
        for edge_row in potential_edge_attributes:
            attribute_number += 1

            attr_edge_id: str = self.make_new_id(edge_id, attribute_number, width)

            kw.writemap({
                node1_column_name: edge_id,
                label_column_name: edge_row[label_column_idx],
                node2_column_name: edge_row[node2_column_idx],
                id_column_name: attr_edge_id
            })
            self.output_line_count += 1
                
            if unreifiedw is not None:
                unreifiedw.writemap({
                    node1_column_name: edge_id,
                    label_column_name: edge_row[label_column_idx],
                    node2_column_name: edge_row[node2_column_idx],
                    id_column_name: attr_edge_id
                })
                

    def pass_group_through(self,
                           kw: KgtkWriter,
                           uninvolvedw: typing.Optional[KgtkWriter],
                           node1_group: typing.List[typing.List[str]],
                           new_id_column: bool):
        # Unreification was not triggered.  Pass this group of rows
        # through unchanged, except for possibly appending an ID
        # column.
        #
        # TODO: Perhaps we'd like to build an ID value at the same time?
        row: typing.List[str]
        for row in node1_group:
            if uninvolvedw is not None:
                uninvolvedw.write(row)
                
            if new_id_column:
                row = row.copy()
                row.append("")

            kw.write(row)
            self.output_line_count += 1
        

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument(      "--trigger-label", dest="trigger_label_value",
                                  help="A value that identifies the trigger label. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_TRIGGER_LABEL_VALUE)

        parser.add_argument(      "--trigger-node2", dest="trigger_node2_value",
                                  help="A value that identifies the trigger node2. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_TRIGGER_NODE2_VALUE)
    
        parser.add_argument(      "--node1-role", dest="rdf_subject_label_value",
                                  help="The label that identifies the edge with the node2 value that will serve in the node1 role. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_RDF_SUBJECT_LABEL_VALUE)
    
        parser.add_argument(      "--label-role", dest="rdf_predicate_label_value",
                                  help="The label that identifies the edge with the node2 value that will serve in the label role. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_RDF_PREDICATE_LABEL_VALUE)
    
        parser.add_argument(      "--node2-role", dest="rdf_object_label_value",
                                  help="The label that identifies the edge with the node2 value that will serve in the node2 role. (default=%(default)s).",
                                  type=str, default=cls.DEFAULT_RDF_OBJECT_LABEL_VALUE)

        parser.add_argument(      "--allow-multiple-subjects", dest="allow_multiple_subjects",
                                  help="When true, allow multiple subjects, resulting in a cartesian product. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_MULTIPLE_SUBJECTS)

        parser.add_argument(      "--allow-multiple-predicates", dest="allow_multiple_predicates",
                                  help="When true, allow multiple predicates, resulting in a cartesian product. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_MULTIPLE_PREDICATES)

        parser.add_argument(      "--allow-multiple-objects", dest="allow_multiple_objects",
                                  help="When true, allow multiple objects, resulting in a cartesian product. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_MULTIPLE_OBJECTS)
            
def main():
    """
    Test the KGTK copy template.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument("-i", "--input-file", dest="input_file_path",
                        help="The KGTK input file. (default=%(default)s)", type=Path, default="-")

    parser.add_argument("-o", "--output-file", dest="output_file_path",
                        help="The KGTK output file. (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--reified-file", dest="reified_file_path",
                              help="A KGTK output file that will contain only the reified RDF statements. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--unreified-file", dest="unreified_file_path",
                              help="A KGTK output file that will contain only the unreified RDF statements. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--uninvolved-file", dest="uninvolved_file_path",
                              help="A KGTK output file that will contain only the uninvolved input records. (default=%(default)s).", type=Path, default=None)
    
    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    KgtkUnreifyRdfStatements.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=False, expert=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("--input-files %s" % " ".join([str(path) for  path in input_file_paths]), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        if args.reified_file_path is not None:
            print("--reified-file=%s" % str(args.reified_file_path), file=error_file, flush=True)
        if args.unreified_file_path is not None:
            print("--unreified-file=%s" % str(args.unreified_file_path), file=error_file, flush=True)
        if args.uninvolved_file_path is not None:
            print("--uninvolved-file=%s" % str(args.uninvolved_file_path), file=error_file, flush=True)

        if args.output_format is not None:
            print("--output-format=%s" % args.output_format, file=error_file, flush=True)

        print("--trigger-label=%s" % args.trigger_label_value, file=error_file, flush=True)
        print("--trigger-node2=%s" % args.trigger_node2_value, file=error_file, flush=True)
        print("--node1-role=%s" % args.rdf_subject_label_value, file=error_file, flush=True)
        print("--label-role=%s" % args.rdf_predicate_label_value, file=error_file, flush=True)
        print("--node2-role=%s" % args.rdf_object_label_value, file=error_file, flush=True)

        print("--allow-multiple-subjects=%s" % str(args.allow_multiple_subjects), file=error_file, flush=True)
        print("--allow-multiple-predicates=%s" % str(args.allow_multiple_predicates), file=error_file, flush=True)
        print("--allow-multiple-objects=%s" % str(args.allow_multiple_objects), file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kurs: KgtkUnreifyRdfStatements = KgtkUnreifyRdfStatements(
        input_file_path=args.input_file_path,
        output_file_path=args.output_file_path,
        reified_file_path=args.reified_file_path,
        unreified_file_path=args.unreified_file_path,
        uninvolved_file_path=args.uninvolved_file_path,

        trigger_label_value=args.trigger_label_value,
        trigger_node2_value=args.trigger_node2_value,
        rdf_object_label_value=args.rdf_object_label_value,
        rdf_predicate_label_value=args.rdf_predicate_label_value,
        rdf_subject_label_value=args.rdf_subject_label_value,

        allow_multiple_subjects=args.allow_multiple_subjects,
        allow_multiple_predicates=args.allow_multiple_predicates,
        allow_multiple_objects=args.allow_multiple_objects,

        reader_options=reader_options,
        value_options=value_options,
        output_format=args.output_format,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose,
    )

    kurs.process()
    
if __name__ == "__main__":
    main()



"""
Join two KTKG edge files or two KGTK node files.  The output file is an edge file or a node file.

Note: This implementation builds in-memory sets of all the key values in
each input file.

"""

from argparse import ArgumentParser
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.join.kgtkmergecolumns import KgtkMergeColumns
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkJoiner(KgtkFormat):
    FIELD_SEPARATOR_DEFAULT: str = KgtkFormat.KEY_FIELD_SEPARATOR

    LEFT: str = "left"
    RIGHT: str = "right"

    left_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    right_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    output_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    # left_join == False and right_join == False: inner join
    # left_join == True and right_join == False: left join
    # left_join == False and right_join == True: right join
    # left_join = True and right_join == True: outer join
    left_join: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    right_join: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # The fllowing may be specified only when both input files are edge files:
    join_on_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    join_on_label: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    join_on_node2: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # TODO: Write fuill validators
    left_join_columns: typing.Optional[typing.List[str]] = attr.ib(default=None)
    right_join_columns: typing.Optional[typing.List[str]] = attr.ib(default=None)

    # The prefix applied to left and right file column names in the output file:
    left_prefix: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    right_prefix: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    # The field separator used in multifield joins.  The KGTK list character should be safe.
    # TODO: USE THE COLUMN SEPARATOR !!!!!
    field_separator: str = attr.ib(validator=attr.validators.instance_of(str), default=KgtkFormat.KEY_FIELD_SEPARATOR)

    # TODO: find working validators:
    left_reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    right_reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def node1_column_idx(self, kr: KgtkReader, who: str)->int:
        idx: int = kr.node1_column_idx
        if idx < 0:
            # TODO: throw a better exception
            raise ValueError("KgtkJoiner: unknown node1 column index in KGTK %s edge file." % who)
        return idx

    def id_column_idx(self, kr: KgtkReader, who: str)->int:
        idx: int = kr.id_column_idx
        if idx < 0:
            # TODO: throw a better exception
            raise ValueError("KgtkJoiner: unknown id column index in KGTK %s node file." % who)
        return idx

    def build_join_key(self, kr: KgtkReader, join_idx_list: typing.List[int], row: typing.List[str])->str:
        key: str = ""
        join_idx: int
        first: bool = True
        for join_idx in join_idx_list:
            if first:
                first = False
            else:
                key += self.field_separator
                
            key += row[join_idx]
        return key

    def multi_column_key_set(self, kr: KgtkReader, join_idx_list: typing.List[int])->typing.Set[str]:
        result: typing.Set[str] = set()
        row: typing.List[str]
        for row in kr:
            result.add(self.build_join_key(kr, join_idx_list, row))
        return result
        
    # Optimized for a single join column:
    def single_column_key_set(self, kr: KgtkReader, join_idx: int)->typing.Set[str]:
        result: typing.Set[str] = set()
        row: typing.List[str]
        for row in kr:
            result.add(row[join_idx])
        return result
        
    def build_join_idx_list(self, kr: KgtkReader, who: str, join_columns: typing.Optional[typing.List[str]])->typing.List[int]:
        join_idx: int
        join_idx_list: typing.List[int] = [ ]
        col_num: int = 1
        if join_columns is not None and len(join_columns) > 0:
            if self.verbose:
                print("Using %s file join columns: %s" % (who, " ".join(join_columns)), file=self.error_file, flush=True)
            join_column:str
            for join_column in join_columns:
                if join_column not in kr.column_name_map:
                    raise ValueError("Join column %s not found in in the %s input file" % (join_column, who))
                join_idx = kr.column_name_map[join_column]
                if self.verbose:
                    print("Join column %d: %s (index %d in the %s input file)" % (col_num, join_column, join_idx, who), file=self.error_file, flush=True)
                join_idx_list.append(join_idx)
            return join_idx_list

        if kr.is_edge_file:
            join_idx = self.node1_column_idx(kr, who)
            if self.verbose:
                print("Joining on node1 (index %s in the %s input file)" % (join_idx, who), file=self.error_file, flush=True)
            join_idx_list.append(join_idx)
        elif kr.is_node_file:
            join_idx = self.id_column_idx(kr, who)
            if self.verbose:
                print("Joining on id (index %s in the %s input file)" % (join_idx, who), file=self.error_file, flush=True)
            join_idx_list.append(join_idx)
        else:
            raise ValueError("Quasi-KGTK files require an explicit list of join columns")

        # join_on_id, join_on_label, and join_on_node2 may be specified
        if self.join_on_id or self.join_on_label or self.join_on_node2:
            if self.join_on_id:
                if kr.id_column_idx < 0:
                    raise ValueError("join_on_id may not be used because the %s input file does not have a id column." % who)
                if self.verbose:
                    print("Joining on id (index %s in the %s input file)" % (kr.id_column_idx, who), file=self.error_file, flush=True)
                join_idx_list.append(kr.id_column_idx)
                
            if self.join_on_label:
                if kr.label_column_idx < 0:
                    raise ValueError("join_on_label may not be used because the %s input file does not have a label column." % who)
                if self.verbose:
                    print("Joining on label (index %s in the %s input file)" % (kr.label_column_idx, who), file=self.error_file, flush=True)
                join_idx_list.append(kr.label_column_idx)
                
            if self.join_on_node2:
                if kr.node2_column_idx < 0:
                    raise ValueError("join_on_node2 may not be used because the %s input file does not have a node2 column." % who)
                if self.verbose:
                    print("Joining on node2 (index %s in the %s input file)" % (kr.node2_column_idx, who), file=self.error_file, flush=True)
                join_idx_list.append(kr.node2_column_idx)
        return join_idx_list
        

    def extract_join_key_set(self, file_path: Path, who: str, join_idx_list: typing.List[int])->typing.Set[str]:
        if self.verbose:
            print("Extracting the join key set from the %s input file: %s" % (who, str(file_path)), file=self.error_file, flush=True)
        reader_options: typing.Optional[KgtkReaderOptions]
        if who == self.LEFT:
            reader_options = self.left_reader_options
        else:
            reader_options = self.right_reader_options
            
        kr: KgtkReader = KgtkReader.open(file_path,
                                         who=who + " input",
                                         options=reader_options,
                                         value_options = self.value_options,
                                         error_file=self.error_file,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        if len(join_idx_list) == 1:
            # This uses optimized code:
            return self.single_column_key_set(kr, join_idx_list[0]) # closes er file
        else:
            return self.multi_column_key_set(kr, join_idx_list) # closes er file
        

    def join_key_sets(self, left_join_idx_list: typing.List[int], right_join_idx_list: typing.List[int])->typing.Optional[typing.Set[str]]:
        """
        Read the input edge files the first time, building the sets of left and right join values.
        """
        join_key_set: typing.Set[str]
        if self.left_join and self.right_join:
            if self.verbose:
                print("Outer join, no need to compute join keys.", file=self.error_file, flush=True)
            return None
        elif self.left_join and not self.right_join:
            if self.verbose:
                print("Computing the left join key set", file=self.error_file, flush=True)
            join_key_set = self.extract_join_key_set(self.left_file_path, self.LEFT, left_join_idx_list).copy()
            if self.verbose:
                print("There are %d keys in the left join key set." % len(join_key_set), file=self.error_file, flush=True)
            return join_key_set

        elif self.right_join and not self.left_join:
            if self.verbose:
                print("Computing the right join key set", file=self.error_file, flush=True)
            join_key_set = self.extract_join_key_set(self.right_file_path, self.RIGHT, right_join_idx_list).copy()
            if self.verbose:
                print("There are %d keys in the right join key set." % len(join_key_set), file=self.error_file, flush=True)
            return join_key_set

        else:
            if self.verbose:
                print("Computing the inner join key set", file=self.error_file, flush=True)
            left_join_key_set: typing.Set[str] = self.extract_join_key_set(self.left_file_path, self.LEFT, left_join_idx_list)
            if self.verbose:
                print("There are %d keys in the left file key set." % len(left_join_key_set), file=self.error_file, flush=True)
            right_join_key_set: typing.Set[str] = self.extract_join_key_set(self.right_file_path, self.RIGHT, right_join_idx_list)
            if self.verbose:
                print("There are %d keys in the right file key set." % len(right_join_key_set), file=self.error_file, flush=True)
            join_key_set = left_join_key_set.intersection(right_join_key_set)
            if self.verbose:
                print("There are %d keys in the inner join key set." % len(join_key_set), file=self.error_file, flush=True)
            return join_key_set
    
    def ok_to_join(self, left_kr: KgtkReader, right_kr: KgtkReader)->bool:
        if left_kr.is_edge_file and right_kr.is_edge_file:
            if self.verbose:
                print("Both input files are edge files.", file=self.error_file, flush=True)
            return True

        elif left_kr.is_node_file and right_kr.is_node_file:
            if self.verbose:
                print("Both input files are node files.", file=self.error_file, flush=True)
            return True

        elif (not (left_kr.is_node_file or left_kr.is_edge_file)) or (not(right_kr.is_edge_file or right_kr.is_node_file)):
            if self.verbose:
               print("One or both input files are quasi-KGTK files.", file=self.error_file, flush=True)
            return True

        else:
            print("Cannot join edge and node files.", file=self.error_file, flush=True)
            return False

    def process(self):
        if self.verbose:
            print("Opening the left edge file: %s" % str(self.left_file_path), file=self.error_file, flush=True)
        left_kr: KgtkReader = KgtkReader.open(self.left_file_path,
                                              who="left input",
                                              options=self.left_reader_options,
                                              value_options = self.value_options,
                                              error_file=self.error_file,
                                              verbose=self.verbose,
                                              very_verbose=self.very_verbose
        )


        if self.verbose:
            print("Opening the right edge file: %s" % str(self.right_file_path), file=self.error_file, flush=True)
        right_kr: KgtkReader = KgtkReader.open(self.right_file_path,
                                               who="right input",
                                               options=self.right_reader_options,
                                               value_options = self.value_options,
                                               error_file=self.error_file,
                                               verbose=self.verbose,
                                               very_verbose=self.very_verbose
        )

        if not self.ok_to_join(left_kr, right_kr):
            left_kr.close()
            right_kr.close()
            return 1

        left_join_idx_list: typing.List[int] = self.build_join_idx_list(left_kr, self.LEFT, self.left_join_columns)
        right_join_idx_list: typing.List[int] = self.build_join_idx_list(right_kr, self.RIGHT, self.right_join_columns)
        if len(left_join_idx_list) != len(right_join_idx_list):
            print("the left join key has %d components, the right join key has %d columns. Exiting." % (len(left_join_idx_list), len(right_join_idx_list)), file=self.error_file, flush=True)
            left_kr.close()
            right_kr.close()
            return 1

        # This might open the input files for a second time. This won't work with stdin.
        joined_key_set: typing.Optional[typing.Set[str]] = self.join_key_sets(left_join_idx_list, right_join_idx_list)

        if self.verbose:
            print("Mapping the column names for the join.", file=self.error_file, flush=True)
        kmc: KgtkMergeColumns = KgtkMergeColumns()
        kmc.merge(left_kr.column_names, prefix=self.left_prefix)
        right_column_names: typing.List[str] = kmc.merge(right_kr.column_names, prefix=self.right_prefix)
        joined_column_names: typing.List[str] = kmc.column_names

        if self.verbose:
            print("       left   columns: %s" % " ".join(left_kr.column_names), file=self.error_file, flush=True)
            print("       right  columns: %s" % " ".join(right_kr.column_names), file=self.error_file, flush=True)
            print("mapped right  columns: %s" % " ".join(right_column_names), file=self.error_file, flush=True)
            print("       joined columns: %s" % " ".join(joined_column_names), file=self.error_file, flush=True)
        
        if self.verbose:
            print("Opening the output edge file: %s" % str(self.output_path), file=self.error_file, flush=True)
        ew: KgtkWriter = KgtkWriter.open(joined_column_names,
                                         self.output_path,
                                         mode=left_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        output_data_lines: int = 0
        left_data_lines_read: int = 0
        left_data_lines_kept: int = 0
        right_data_lines_read: int = 0
        right_data_lines_kept: int = 0
        
        if self.verbose:
            print("Processing the left input file: %s" % str(self.left_file_path), file=self.error_file, flush=True)
        row: typing.List[str]
        for row in left_kr:
            left_data_lines_read += 1
            if joined_key_set is None:
                ew.write(row)
                output_data_lines += 1
                left_data_lines_kept += 1
            else:
                left_key: str = self.build_join_key(left_kr, left_join_idx_list, row)
                if left_key in joined_key_set:
                    ew.write(row)
                    output_data_lines += 1
                    left_data_lines_kept += 1
        # Flush the output file so far:
        ew.flush()

        if self.verbose:
            print("Processing the right input file: %s" % str(self.right_file_path), file=self.error_file, flush=True)
        right_shuffle_list: typing.List[int] = ew.build_shuffle_list(right_column_names)
        for row in right_kr:
            right_data_lines_read += 1
            if joined_key_set is None:
                ew.write(row, shuffle_list=right_shuffle_list)
                output_data_lines += 1
                right_data_lines_kept += 1
            else:
                right_key: str = self.build_join_key(right_kr, right_join_idx_list, row)
                if right_key in joined_key_set:
                    ew.write(row, shuffle_list=right_shuffle_list)
                    output_data_lines += 1
                    right_data_lines_kept += 1
            
        ew.close()
        if self.verbose:
            print("The join is complete", file=self.error_file, flush=True)
            print("%d left input data lines read, %d kept" % (left_data_lines_read, left_data_lines_kept), file=self.error_file, flush=True)
            print("%d right input data lines read, %d kept" % (right_data_lines_read, right_data_lines_kept), file=self.error_file, flush=True)
            print("%d data lines written." % output_data_lines, file=self.error_file, flush=True)
        
def main():
    """
    Test the KGTK file joiner.

    Edge files can be joined to edge files.
    Node files can be joined to node files.

    TODO: Add more KgtkReader parameters, especially mode.
    """
    parser = ArgumentParser()
    parser.add_argument(dest="left_file_path", help="The left KGTK file to join", type=Path)
    parser.add_argument(dest="right_file_path", help="The right KGTK file to join", type=Path)
    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write", type=Path, default=None)

    parser.add_argument(      "--field-separator", dest="field_separator", help="Separator for multifield keys", default=KgtkJoiner.FIELD_SEPARATOR_DEFAULT)

    parser.add_argument(      "--join-on-id", dest="join_on_id",
                              help="If both input files are edge files, include the id column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    parser.add_argument(      "--join-on-label", dest="join_on_label",
                              help="If both input files are edge files, include the label column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    parser.add_argument(      "--join-on-node2", dest="join_on_node2",
                              help="If both input files are edge files, include the node2 column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    parser.add_argument(      "--left-prefix", dest="left_prefix", help="An optional prefix applied to left file column names in the output file (default=None).")
    parser.add_argument(      "--left-join", dest="left_join", help="Perform a left outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    parser.add_argument(      "--left-file-join-columns", dest="left_join_columns", help="Left file join columns.", nargs='+')

    parser.add_argument(      "--right-prefix", "--prefix", dest="right_prefix", help="An optional prefix applied to right file column names in the output file (default=None).")
    parser.add_argument(      "--right-file-join-columns", dest="right_join_columns", help="Right file join columns.", nargs='+')
    parser.add_argument(      "--right-join", dest="right_join", help="Perform a right outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=True)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who=KgtkJoiner.LEFT, expert=True)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who=KgtkJoiner.RIGHT, expert=True)
    KgtkValueOptions.add_arguments(parser, expert=True)

    args = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.
    left_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args, who=KgtkJoiner.LEFT)
    right_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args, who=KgtkJoiner.RIGHT)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        if args.left_prefix is not None:
            print("--left-prefix=%s" % args.left_prefix, file=error_file, flush=True)
        if args.right_prefix is not None:
            print("--right-prefix=%s" % args.right_prefix, file=error_file, flush=True)
        left_reader_options.show(out=error_file, who=KgtkJoiner.LEFT)
        right_reader_options.show(out=error_file, who=KgtkJoiner.RIGHT)
        value_options.show(out=error_file)

    ej: KgtkJoiner = KgtkJoiner(left_file_path=args.left_file_path,
                                right_file_path=args.right_file_path,
                                output_path=args.output_file_path,
                                left_join=args.left_join,
                                right_join=args.right_join,
                                join_on_id=args.join_on_id,
                                join_on_label=args.join_on_label,
                                join_on_node2=args.join_on_node2,
                                left_join_columns=args.left_join_columns,
                                right_join_columns=args.right_join_columns,
                                left_prefix=args.left_prefix,
                                right_prefix=args.right_prefix,
                                field_separator=args.field_separator,
                                left_reader_options=left_reader_options,
                                right_reader_options=right_reader_options,
                                value_options=value_options,
                                error_file=error_file,
                                verbose=args.verbose,
                                very_verbose=args.very_verbose)

    ej.process()

if __name__ == "__main__":
    main()




"""
Concatenate KGTK files.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Concatenate KGTK files.',
        'description': 'Concatenate two or more KGTK files, merging the columns appropriately. ' +
        'All files must be KGTK edge files or all files must be KGTK node files (unless overridden with --mode=NONE). ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert cat --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="KGTK input files",
                          dest="input_files",
                          options=["-i", "--input-files"],
                          allow_list=True,
                          positional=True)
    parser.add_output_file()

    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    parser.add_argument(      "--output-columns", dest="output_column_names",
                              metavar="NEW_COLUMN_NAME",
                              help=h("The list of new column names when renaming all columns."),
                              type=str, nargs='+')
    parser.add_argument(      "--old-columns", dest="old_column_names",
                              metavar="OLD_COLUMN_NAME",
                              help=h("The list of old column names for selective renaming."),
                              type=str, nargs='+')
    parser.add_argument(      "--new-columns", dest="new_column_names",
                              metavar="NEW_COLUMN_NAME",
                              help=h("The list of new column names for selective renaming."),
                              type=str, nargs='+')

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_files: KGTKFiles,
        output_file: KGTKFiles,
        output_format: typing.Optional[str],

        output_column_names: typing.Optional[typing.List[str]],
        old_column_names: typing.Optional[typing.List[str]],
        new_column_names: typing.Optional[typing.List[str]],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions
    
    input_file_paths: typing.List[Path] = KGTKArgumentParser.get_input_file_list(input_files)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)
    
    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # print("cat running", file=error_file, flush=True) # ***

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-files %s" % " ".join((str(input_file_path) for input_file_path in input_file_paths)), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format=%s" % output_format, file=error_file, flush=True)
        if output_column_names is not None:
            print("--output-coloumns %s" % " ".join(output_column_names), file=error_file, flush=True)
        if old_column_names is not None:
            print("--old-columns %s" % " ".join(old_column_names), file=error_file, flush=True)
        if new_column_names is not None:
            print("--new-columns %s" % " ".join(new_column_names), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Check for comsistent options.  argparse doesn't support this yet.
    if output_column_names is not None and len(output_column_names) > 0:
        if (old_column_names is not None and len(old_column_names) > 0) or \
           (new_column_names is not None and len(new_column_names) > 0):
            raise KGTKException("When --output-columns is used, --old-columns and --new-columns may not be used.")
    elif (old_column_names is not None and len(old_column_names) > 0) ^ \
         (new_column_names is not None and len(new_column_names) > 0):
        raise KGTKException("Both --old-columns and --new-columns must be used when either is used.")

    elif (old_column_names is not None and len(old_column_names) > 0) and \
         (new_column_names is not None and len(new_column_names) > 0):
        if len(old_column_names) != len(new_column_names):
            raise KGTKException("Both --old-columns and --new-columns must have the same number of columns.")
    try:
        kc: KgtkCat = KgtkCat(input_file_paths=input_file_paths,
                              output_path=output_file_path,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names,
                              new_column_names=new_column_names,
                              reader_options=reader_options,
                              value_options=value_options,
                              error_file=error_file,
                              verbose=verbose,
                              very_verbose=very_verbose
        )
        
        kc.process()

        # print("cat done", file=error_file, flush=True) # ***
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Example CLI module

Please DON'T import specific modules globally, import them in `run`.
Please DON'T initialize resource (e.g., variable) globally.
"""
from kgtk.cli_argparse import KGTKArgumentParser


def parser():
    """
    Initialize sub-parser.
    Parameters: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser
    """
    return {
        'help': 'this is example',
        'description': 'this is a basic example'
    }


def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
        parser (kgtk.cli_argparse.KGTKArgumentParser)
    """
    parser.add_argument(action="store", type=str, metavar="name", dest="name")
    parser.add_argument("-i", "--info", action="store", type=str, dest="info")
    parser.add_argument("-e", "--error", action="store_true", help="raise an error")
    parser.accept_shared_argument('_debug')


def run(name, info, error, _debug):
    """
    Arguments here should be defined in `add_arguments` first.
    The return value (integer) will be the return code in shell. It will set to 0 if no value returns.
    Though you can return a non-zero value to indicate error, raise exceptions defined in kgtk.exceptions is preferred
    since this gives user an unified error code and message.
    """
    # import modules locally
    import socket
    from kgtk.exceptions import KGTKException

    if _debug:
        print('DEBUG MODE')

    if error:
        raise KGTKException('An error here\n')

    print('name: {}, info: {}\nhost: {}'.format(name, info, socket.gethostname()))



"""Copy records from the first KGTK file to the output file,
expanding | lists.

TODO: Need KgtkWriterOptions

TODO: --columns should list the columns to expand. 
      --key-columns should force the key columns.
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Copy a KGTK file expanding | lists.',
        'description': 'Copy a KGTK file, expanding | lists into multiple records. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert expand --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument(      "--columns", dest="key_column_names",
                              help="The key columns will not be expanded.  They will be repeated on each output record. " +
                              "(default=id for node files, (node1, label, node2) for edge files).", nargs='+', default=[ ])


    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        key_column_names: typing.List[str],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkexpand import KgtkExpand
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        print("--columns=%s" % " ".join(key_column_names), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ex: KgtkExpand = KgtkExpand(
            input_file_path=input_kgtk_file,
            key_column_names=key_column_names,
            output_file_path=output_kgtk_file,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))

